# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=redefined-builtin, arguments-differ, line-too-long

"""Friendli Catalog SDK."""

from __future__ import annotations

from typing import List, Optional
from uuid import UUID

from friendli.client.catalog import CatalogClient
from friendli.schema.resource.v1.checkpoint import V1Catalog
from friendli.sdk.resource.base import ResourceAPI
from friendli.utils.fs import strip_storage_path_prefix


class Catalog(ResourceAPI[V1Catalog, UUID]):
    """Catalog resource API."""

    def create(self, *args, **kwargs) -> V1Catalog:
        """[skip-doc]."""
        raise NotImplementedError

    def list(self, *, name: Optional[str] = None, limit: int = 20) -> List[V1Catalog]:
        """Lists public checkpoints in the catalog.

        Args:
            name (Optional[str], optional): The name of public checkpoint as a search key. Defaults to None.
            limit (int, optional): The maximum number of retrieved results. Defaults to 20.

        Returns:
            List[V1Catalog]: A list of retrieved public checkpoints from the catalog.

        """
        client = CatalogClient()
        catalogs = [
            V1Catalog.model_validate(raw_catalog)
            for raw_catalog in client.list_catalogs(name=name, limit=limit)
        ]
        for catalog in catalogs:
            for file in catalog.files:
                file.path = strip_storage_path_prefix(file.path)
        return catalogs

    def get(self, id: UUID, *args, **kwargs) -> V1Catalog:
        """Get a catalog info.

        Args:
            id (UUID): ID of a catalog.

        Returns:
            V1Catalog: The retrieved catalog object.

        """
        client = CatalogClient()
        raw_catalog = client.get_catalog(catalog_id=id)
        catalog = V1Catalog.model_validate(raw_catalog)
        for file in catalog.files:
            file.path = strip_storage_path_prefix(file.path)
        return catalog



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=too-many-locals, redefined-builtin, too-many-statements, too-many-arguments

"""Friendli Deployment CLI."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Optional
from uuid import UUID

import typer
import yaml
from dateutil.parser import parse

from friendli.client.user import UserGroupProjectClient
from friendli.enums import CloudType, DeploymentSecurityLevel, DeploymentType, GpuType
from friendli.errors import (
    AuthenticationError,
    EntityTooLargeError,
    InvalidConfigError,
    LowServicePlanError,
)
from friendli.formatter import PanelFormatter, TableFormatter
from friendli.sdk.resource.deployment import Deployment
from friendli.utils.decorator import check_api
from friendli.utils.format import (
    datetime_to_pretty_str,
    datetime_to_simple_string,
    secho_error_and_exit,
)

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

DEFAULT_MAX_BATCH_SIZE = 256
DEFAULT_MAX_TOKEN_COUNT = 8192

deployment_panel = PanelFormatter(
    name="Deployment Overview",
    fields=[
        "deployment_id",
        "config.name",
        "description",
        "config.deployment_type",
        "config.model_id",
        "status",
        "ready_replicas",
        "config.scaler_config.min_replica_count",
        "config.scaler_config.max_replica_count",
        "vms",
        "config.vm.gpu_type",
        "config.total_gpus",
        "start",
        "end",
        "security_level",
        "config.infrequest_log",
        "endpoint",
        "config.cloud",
        "config.region",
        "config.orca_config.max_batch_size",
        "config.orca_config.max_token_count",
        "config.orca_config.max_num_tokens_to_replace",
    ],
    headers=[
        "ID",
        "Name",
        "Description",
        "Type",
        "Ckpt ID",
        "Status",
        "#Ready",
        "Min Replicas",
        "Max Replicas",
        "VM Type",
        "GPU Type",
        "#GPUs",
        "Start",
        "End",
        "Security Level",
        "Logging",
        "Endpoint",
        "Cloud",
        "Region",
        "Max batch size",
        "Max token count",
        "Max num tokens to replace",
    ],
    extra_fields=["error"],
    extra_headers=["error"],
    substitute_exact_match_only=False,
)

deployment_table = TableFormatter(
    name="Deployments",
    fields=[
        "deployment_id",
        "config.name",
        "status",
        "ready_replicas",
        "vms",
        "config.vm.gpu_type",
        "config.total_gpus",
        "start",
        "config.cloud",
        "config.region",
    ],
    headers=[
        "ID",
        "Name",
        "Status",
        "#Ready",
        "VM Type",
        "GPU Type",
        "#GPUs",
        "Start",
        "Cloud",
        "Region",
    ],
    extra_fields=["error"],
    extra_headers=["error"],
    substitute_exact_match_only=False,
)

deployment_org_table = TableFormatter(
    name="Deployments",
    fields=[
        "deployment_id",
        "config.name",
        "status",
        "ready_replicas",
        "vms",
        "config.vm.gpu_type",
        "config.total_gpus",
        "start",
        "config.cloud",
        "config.region",
        "config.project_id",
        "project_name",
    ],
    headers=[
        "ID",
        "Name",
        "Status",
        "#Ready",
        "VM Type",
        "GPU Type",
        "#GPUs",
        "Start",
        "Cloud",
        "Region",
        "Project ID",
        "Project Name",
    ],
    extra_fields=["error"],
    extra_headers=["error"],
    substitute_exact_match_only=False,
)

deployment_usage_table = TableFormatter(
    name="Deployment Usage",
    fields=[
        "id",
        "type",
        "cloud",
        "vm",
        "created_at",
        "finished_at",
        "gpu_type",
        "duration",
    ],
    headers=[
        "ID",
        "Type",
        "Cloud",
        "VM",
        "Created At",
        "Finished At",
        "GPU",
        "Total Usage (days, HH:MM:SS)",
    ],
)

deployment_event_table = TableFormatter(
    name="Deployment Event",
    fields=[
        "id",
        "type",
        "description",
        "created_at",
    ],
    headers=["ID", "Type", "Description", "Timestamp"],
)

deployment_panel.add_substitution_rule(
    "Initializing", "[bold yellow]Initializing[/bold yellow]"
)
deployment_panel.add_substitution_rule("Healthy", "[bold green]Healthy[/bold green]")
deployment_panel.add_substitution_rule("Unhealthy", "[bold red]Unhealthy[/bold red]")
deployment_panel.add_substitution_rule(
    "Stopping", "[bold magenta]Stopping[/bold magenta]"
)
deployment_panel.add_substitution_rule("Failed", "[bold red]Failed[/bold red]")
deployment_panel.add_substitution_rule("Terminated", "[bold]Terminated[/bold]")

deployment_table.add_substitution_rule(
    "Initializing", "[bold yellow]Initializing[/bold yellow]"
)
deployment_table.add_substitution_rule("Healthy", "[bold green]Healthy[/bold green]")
deployment_table.add_substitution_rule("Unhealthy", "[bold red]Unhealthy[/bold red]")
deployment_table.add_substitution_rule(
    "Stopping", "[bold magenta]Stopping[/bold magenta]"
)
deployment_table.add_substitution_rule("Failed", "[bold red]Failed[/bold red]")
deployment_table.add_substitution_rule("Terminated", "[bold]Terminated[/bold]")

deployment_org_table.add_substitution_rule(
    "Initializing", "[bold yellow]Initializing[/bold yellow]"
)
deployment_org_table.add_substitution_rule(
    "Healthy", "[bold green]Healthy[/bold green]"
)
deployment_org_table.add_substitution_rule(
    "Unhealthy", "[bold red]Unhealthy[/bold red]"
)
deployment_org_table.add_substitution_rule(
    "Stopping", "[bold magenta]Stopping[/bold magenta]"
)
deployment_org_table.add_substitution_rule("Failed", "[bold red]Failed[/bold red]")
deployment_org_table.add_substitution_rule("Terminated", "[bold]Terminated[/bold]")


def get_deployment_id_from_namespace(namespace: str):
    """Get deployment id from namespace."""
    return f"friendli-deployment-{namespace}"


@app.command()
@check_api
def list(
    include_terminated: bool = typer.Option(
        False,
        "--include-terminated",
        help=(
            "When True, shows all deployments in my project including terminated "
            "deployments. (active deployments are shown above the terminated ones.)"
        ),
    ),
    limit: int = typer.Option(20, "--limit", help="The number of deployments to view"),
    from_oldest: bool = typer.Option(
        False, "--from-oldest", help="When True, shows from the oldest deployments."
    ),
    org: bool = typer.Option(False, "--org", help="Shows all deployments in org"),
):
    """Lists all deployments."""
    client = Deployment()
    try:
        deployments = client.list(
            limit=limit,
            include_terminated=include_terminated,
            from_oldest=from_oldest,
            all_in_org=org,
        )
    except AuthenticationError as exc:
        secho_error_and_exit(str(exc))

    deployment_dicts = []
    for deployment in deployments:
        deployment_dict = deployment.model_dump()
        deployment_dict["start"] = datetime_to_pretty_str(deployment.start)
        deployment_dict["vms"] = deployment.vms[0].name if deployment.vms else "None"
        deployment_dicts.append(deployment_dict)

    table = deployment_table
    if org:
        table = deployment_org_table
        project_client = UserGroupProjectClient()
        projects = project_client.list_projects()
        project_map = {project["id"]: project["name"] for project in projects}
        for deployment_dict in deployment_dicts:
            project_id = str(deployment_dict["config"]["project_id"])
            deployment_dict["project_name"] = (
                project_map[project_id] if project_id in project_map else project_id
            )

    table.render(deployment_dicts)


@app.command()
@check_api
def stop(deployment_id: str = typer.Argument(..., help="ID of deployment to stop")):
    """Stops a running deployment."""
    client = Deployment()
    client.stop(id=deployment_id)
    typer.secho(
        f"Deployment ({deployment_id}) stopped successfully.",
        fg=typer.colors.GREEN,
    )


@app.command()
@check_api
def view(
    deployment_id: str = typer.Argument(..., help="deployment id to inspect detail.")
):
    """Shows details of a deployment."""
    client = Deployment()
    deployment = client.get(id=deployment_id)

    deployment_dict = deployment.model_dump()
    deployment_dict["start"] = deployment.start.ctime()
    deployment_dict["end"] = deployment.end and deployment.end.ctime()
    deployment_dict["vms"] = deployment.vms[0].name if deployment.vms else None
    deployment_dict["security_level"] = (
        DeploymentSecurityLevel.PROTECTED
        if deployment.config.infrequest_perm_check
        else DeploymentSecurityLevel.PUBLIC
    )
    deployment_panel.render([deployment_dict])


@app.command()
@check_api
def metrics(
    deployment_id: str = typer.Argument(
        ..., help="ID of deployment to inspect in detail."
    ),
    since: str = typer.Option(
        ...,
        "--since",
        help=(
            "Start time of metrics to fetch. The format should be {YYYY}-{MM}-{DD}T{HH}. "
            "The UTC timezone will be used by default."
        ),
    ),
    until: str = typer.Option(
        ...,
        "--until",
        help=(
            "End time of metrics to fetch. The format should be {YYYY}-{MM}-{DD}T{HH}. "
            "The UTC timezone will be used by default."
        ),
    ),
    window: int = typer.Option(
        60, "--window", "-w", help="Time window of metrics in seconds."
    ),
):
    """Show metrics of a deployment."""
    try:
        start = datetime.strptime(since, "%Y-%m-%dT%H").astimezone(tz=timezone.utc)
        end = datetime.strptime(until, "%Y-%m-%dT%H").astimezone(tz=timezone.utc)
    except ValueError:
        secho_error_and_exit(
            "Invalid datetime format. The format should be {YYYY}-{MM}-{DD}T{HH} "
            "(e.g., 1999-01-01T01)."
        )

    client = Deployment()
    metrics = client.get_metrics(
        id=deployment_id, since=start, until=end, time_window=window
    )
    for metric in metrics:
        metric_key = metric["metric"]
        metric_data = metric["points"]

        deployment_metrics_table = TableFormatter(
            name=metric_key,
            fields=[
                "value",
                "timestamp",
            ],
            headers=["Value", "Timestamp"],
        )
        deployment_metrics_table.render(metric_data)


@app.command()
@check_api
def usage(
    year: int = typer.Argument(...),
    month: int = typer.Argument(...),
    day: Optional[int] = typer.Argument(None),
):
    """Shows the usage of all deployments in a project within a specific month or day.

    :::info
    - Timestamps are recorded in UTC.
    - The usages are updated every minute.
    :::

    """
    try:
        start_date = datetime(year, month, day if day else 1, tzinfo=timezone.utc)
    except ValueError:
        secho_error_and_exit(f"Invalid date({year}-{month}{f'-{day}' if day else ''})")
    if day:
        end_date = start_date + timedelta(days=1)
    else:
        end_date = datetime(
            year + int(month == 12),
            (month + 1) if month < 12 else 1,
            1,
            tzinfo=timezone.utc,
        )

    client = Deployment()
    usages = client.get_project_deployment_durations(since=start_date, until=end_date)
    deployments = [
        {
            "id": id,
            "type": info["deployment_type"],
            "cloud": info["cloud"].upper() if "cloud" in info else None,
            "vm": info["vm"]["name"] if info.get("vm") else None,
            "gpu_type": info["vm"]["gpu_type"].upper() if info.get("vm") else None,
            "created_at": datetime_to_simple_string(parse(info["created_at"])),
            "finished_at": datetime_to_simple_string(parse(info["finished_at"]))
            if info["finished_at"]
            else "-",
            "duration": timedelta(seconds=int(info["duration"])),
        }
        for id, info in usages.items()
        if int(info["duration"]) != 0
    ]
    deployment_usage_table.render(deployments)


@app.command()
@check_api
def log(
    deployment_id: str = typer.Argument(..., help="ID of deployment to get log."),
    replica_index: int = typer.Argument(
        0, help="Replica index of deployment to get log."
    ),
):
    """Shows deployments log."""
    client = Deployment()
    logs = client.get_logs(id=deployment_id, replica_index=replica_index)

    if len(logs) == 0:
        secho_error_and_exit("Logs are not found yet.")

    for line in logs:
        typer.echo(line["data"])


@app.command()
@check_api
def create(
    checkpoint_id: UUID = typer.Option(
        ..., "--checkpoint-id", "-i", help="Checkpoint id to deploy."
    ),
    deployment_name: str = typer.Option(
        ..., "--name", "-n", help="The name of deployment. "
    ),
    cloud: CloudType = typer.Option(..., "--cloud", "-c", help="Type of cloud."),
    region: str = typer.Option(..., "--region", "-r", help="Region of cloud."),
    gpu_type: GpuType = typer.Option(
        ..., "--gpu-type", "-g", help="The GPU type for the deployment."
    ),
    num_gpus: int = typer.Option(
        ...,
        "--num-gpus",
        "-ng",
        help="The number of GPUs for the deployment. Equals to the tensor parallelism degree.",
    ),
    config_file: Optional[typer.FileText] = typer.Option(
        None, "--config-file", "-f", help="Path to configuration file."
    ),
    deployment_type: DeploymentType = typer.Option(
        DeploymentType.PROD, "--type", "-t", help="Type of deployment."
    ),
    description: Optional[str] = typer.Option(
        None, "--description", "-d", help="Deployment description."
    ),
    security_level: DeploymentSecurityLevel = typer.Option(
        DeploymentSecurityLevel.PUBLIC.value,
        "--security-level",
        "-sl",
        help="Security level of deployment endpoints",
    ),
    logging: bool = typer.Option(
        False,
        "--logging",
        "-l",
        help="Logging inference requests or not.",
    ),
    default_request_config_file: Optional[typer.FileText] = typer.Option(
        None,
        "--default-request-config-file",
        "-drc",
        help="Path to default request config",
    ),
    min_replicas: int = typer.Option(
        1,
        "--min-replicas",
        "-min",
        help="Number of minimum replicas.",
    ),
    max_replicas: int = typer.Option(
        1,
        "--max-replicas",
        "-max",
        help="Number of maximum replicas.",
    ),
):
    """Creates a deployment object by using model checkpoint.

    The deployment settings are described in a configuration YAML file, and the path of that file is
    passed to the `-f` option. The following is an example YAML file:

    ```yaml
    max_batch_size: 384
    max_token_count: 12288
    max_num_tokens_to_replace: 0
    ```

    :::tip
    To turn off the deployment autoscaling, set `--min-replicas` and
    `--max-replicas` to the same value.
    :::

    :::tip
    Use `friendli vm list` to find available vm-type, cloud, region, and gpu-type.
    :::

    The default request-response configuration, such as stop tokens or bad words, is
    defined in a YAML file. The path of that file is passed to the `-drc` option.
    The format of the file is as follows.

    ```json
    {
        "stop": Optional[List[str]],
        "stop_tokens": Optional[List[TokenSequence]],
        "bad_words": Optional[List[str]],
        "bad_word_tokens": Optional[List[TokenSequence]]
    }
    ```

    :::caution
    Both `bad_words` and `bad_word_tokens` cannot be set at the same time. Also, both
    `stop` and `stop_tokens` cannot be set at the same time.

    And a TokenSequence type is a dict with the key 'tokens' and the value type List[int].
    ```json
    # Wrong Example: Both `bad_words` and `bad_word_tokens` are set.
    {
        "bad_words": ["bad", "words", " bad", " words"]
        "bad_word_tokens": [{"tokens": [568, 36]}, {"tokens": [423, 76]}]
    }

    # Correct Example
    {
        "bad_words": ["bad", "words", " bad", " words"]
        "stop_tokens": [{"tokens": [568, 36]}, {"tokens": [423, 76]}]
    }
    :::

    """
    default_request_config = None
    config: Dict[str, Any] = {}
    if config_file:
        try:
            config = yaml.safe_load(config_file)
        except yaml.YAMLError as e:
            secho_error_and_exit(
                f"Error occurred while parsing engine config file... {e}"
            )

    if "max_batch_size" not in config:
        config["max_batch_size"] = DEFAULT_MAX_BATCH_SIZE
    if "max_token_count" not in config:
        config["max_token_count"] = DEFAULT_MAX_TOKEN_COUNT

    if default_request_config_file is not None:
        try:
            default_request_config = yaml.safe_load(default_request_config_file)
        except yaml.YAMLError as e:
            secho_error_and_exit(
                f"Error occurred while parsing default request config file... {e}"
            )

    client = Deployment()
    try:
        deployment = client.create(
            checkpoint_id=checkpoint_id,
            name=deployment_name,
            deployment_type=deployment_type,
            cloud=cloud,
            region=region,
            gpu_type=gpu_type,
            num_gpus=num_gpus,
            config=config,
            description=description,
            default_request_config=default_request_config,
            security_level=security_level,
            logging=logging,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
        )
    except (
        AuthenticationError,
        InvalidConfigError,
        EntityTooLargeError,
        LowServicePlanError,
    ) as exc:
        secho_error_and_exit(str(exc))

    typer.secho(
        f"Deployment ({deployment.deployment_id}) started successfully. "
        f"Use 'friendli deployment view {deployment.deployment_id}' to see details.\n"
        f"Send inference requests to '{deployment.endpoint}'.",
        fg=typer.colors.GREEN,
    )


@app.command()
@check_api
def update(
    deployment_id: str = typer.Argument(..., help="ID of deployment to update."),
    min_replicas: int = typer.Option(
        ..., "--min-replicas", "-min", help="Set min_replicas of deployment."
    ),
    max_replicas: int = typer.Option(
        ..., "--max-replicas", "-max", help="Set max_replicas of deployment."
    ),
):
    """Updates configuration of a deployment.

    :::tip
    To turn off the deployment autoscaling, set `--min-replicas` and `--max-replicas`
    to the same value.
    :::

    """
    client = Deployment()
    client.adjust_replica_config(
        id=deployment_id, min_replicas=min_replicas, max_replicas=max_replicas
    )
    typer.secho(
        f"Scaler of deployment ({deployment_id}) is updated.\n"
        f"Set min_replicas to {min_replicas}, max_replicas to {max_replicas}",
        fg=typer.colors.GREEN,
    )


@app.command()
@check_api
def event(
    deployment_id: str = typer.Argument(..., help="ID of deployment to get events."),
):
    """Gets deployment events."""
    client = Deployment()
    events = client.get_events(id=deployment_id)
    for event in events:
        event["id"] = f"friendli-deployment-{event['namespace']}"
        event["created_at"] = datetime_to_simple_string(parse(event["created_at"]))
    deployment_event_table.render(events)


@app.command()
@check_api
def req_resp(
    deployment_id: str = typer.Argument(
        ..., help="ID of deployment to download request-response logs."
    ),
    since: str = typer.Option(
        ...,
        "--since",
        help=(
            "Start time of request-response logs. The format should be {YYYY}-{MM}-{DD}T{HH}. "
            "The UTC timezone will be used by default."
        ),
    ),
    until: str = typer.Option(
        ...,
        "--until",
        help=(
            "End time of request-response logs. The format should be {YYYY}-{MM}-{DD}T{HH}. "
            "The UTC timezone will be used by default."
        ),
    ),
    save_dir: Optional[str] = typer.Option(
        None,
        "-s",
        "--save-dir",
        help="Directory path to save request-response logs",
    ),
):
    """Downloads request-response logs for a deployment."""
    try:
        start = datetime.strptime(since, "%Y-%m-%dT%H")
        end = datetime.strptime(until, "%Y-%m-%dT%H")
    except ValueError:
        secho_error_and_exit(
            "Invalid datetime format. The format should be {YYYY}-{MM}-{DD}T{HH} "
            "(e.g., 1999-01-01T01)."
        )

    client = Deployment()
    client.download_req_resp_logs(
        id=deployment_id, since=start, until=end, save_dir=save_dir
    )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Logger."""

from __future__ import annotations

import logging
import os

_formatter = logging.Formatter(
    fmt="%(asctime)s.%(msecs)05d: %(name)s %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def get_logger(name: str) -> logging.Logger:
    """Get a formatted logger."""
    logger = logging.getLogger(name)

    handler = logging.StreamHandler()
    handler.setFormatter(_formatter)
    logger.addHandler(handler)
    logger.setLevel(os.environ.get("FRIENDLI_LOG_LEVEL", "INFO"))

    return logger


logger = get_logger("Friendli")



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli client enums."""


from __future__ import annotations

from enum import Enum


class GroupRole(str, Enum):
    """Organization-level roles."""

    OWNER = "owner"
    MEMBER = "member"


class ProjectRole(str, Enum):
    """Project-level roles."""

    ADMIN = "admin"
    MAINTAIN = "maintain"
    DEVELOP = "develop"
    GUEST = "guest"


class ServiceTier(str, Enum):
    """Organization service tier."""

    BASIC = "basic"
    ENT = "enterprise"


class CloudType(str, Enum):
    """Cloud types."""

    AWS = "aws"
    AZURE = "azure"
    GCP = "gcp"


class StorageType(str, Enum):
    """Cloud storage types."""

    S3 = "s3"
    BLOB = "azure-blob"
    GCS = "gcs"
    FAI = "fai"


class GpuType(str, Enum):
    """GPU types for deployment."""

    A10G = "a10g"
    A100 = "a100"
    A100_80G = "a100-80g"


class VMType(str, Enum):
    """VM types for deployment."""

    G5_XLARGE = "g5.xlarge"
    A2_HIGHGPU_1G = "a2-highgpu-1g"
    A2_ULTRAGPU_1G = "a2-ultragpu-1g"
    A2_ULTRAGPU_2G = "a2-ultragpu-2g"
    A2_ULTRAGPU_4G = "a2-ultragpu-4g"
    A2_ULTRAGPU_8G = "a2-ultragpu-8g"


class DeploymentType(str, Enum):
    """Deployment phase types."""

    DEV = "dev"
    PROD = "prod"


class DeploymentSecurityLevel(str, Enum):
    """Deployment access levels."""

    PUBLIC = "public"
    PROTECTED = "protected"


class CheckpointCategory(str, Enum):
    """Checkpoint categories."""

    USER_PROVIDED = "USER"
    JOB_GENERATED = "JOB"
    COPIED = "COPY"
    REFERENCED = "REF"


class CheckpointStatus(str, Enum):
    """Checkpoint statuses."""

    CREATED = "Created"  # Model is created
    COPYING = "Copying"  # Copying the model from catalog
    ACTIVE = "Active"  # Available to use
    FAILED = "Failed"  # Cannot use the checkpoint


class CheckpointValidationStatus(str, Enum):
    """Checkpoint validation statuses."""

    VALIDATING = "Validating"  # Validating the checkpoint format
    VALID = "Valid"  # Checkpoint format is valid
    INVALID = "Invalid"  # Checkpoint is invalid
    FAILED = "Failed"  # Validation failed for some reason


class CatalogImportMethod(str, Enum):
    """Catalog import methods."""

    COPY = "copy"
    REF = "ref"


class CatalogStatus(str, Enum):
    """Catalog statuses."""

    CREATED = "Created"  # Catalog is created
    PUBLISHING = "Publishing"  # Publishing the catalog
    ACTIVE = "Active"  # Available to use the catalog
    FAILED = "Failed"  # Failed to publish catalog


class CredType(str, Enum):
    """Credential types."""

    S3 = "s3"
    BLOB = "azure-blob"
    GCS = "gcs"


class DeploymentReplicaStatus(str, Enum):
    """Deployment replica status types."""

    INITIALIZING = "Initializing"
    PENDING = "Pending"
    RUNNING = "Running"
    SUCCESS = "Success"
    FAILED = "Failed"
    UNKNOWN = "Unknown"
    STOPPING = "Stopping"
    TERMINATED = "Terminated"


class DeploymentStatus(str, Enum):
    """Deployment status types."""

    INITIALIZING = "Initializing"
    HEALTHY = "Healthy"
    UNHEALTHY = "Unhealthy"
    FAILED = "Failed"
    STOPPING = "Stopping"
    TERMINATED = "Terminated"


class CheckpointDataType(str, Enum):
    """Checkpoint dtype enums."""

    BF16 = "bf16"
    FP16 = "fp16"
    FP32 = "fp32"
    INT8 = "int8"
    INT4 = "int4"


class CheckpointFileType(str, Enum):
    """Checkpoint file types."""

    HDF5 = "hdf5"
    SAFETENSORS = "safetensors"

    def __str__(self):
        """Convert to a human-readable string."""
        return self.value


class FileSizeType(str, Enum):
    """File size type."""

    LARGE = "LARGE"
    SMALL = "SMALL"


class QuantMode(str, Enum):
    """Supported quantization modes."""

    SMOOTH_QUANT = "smoothquant"
    AWQ = "awq"


class QuantDatasetFormat(str, Enum):
    """Supported file format for calibration datasets for quantization."""

    JSON = "json"
    CSV = "csv"
    PARQUET = "parquet"
    TXT = "txt"


class ChatCompletionsModel(str, Enum):
    """Model choices for chat completions."""

    LLAMA_2_13B_CHAT = "llama-2-13b-chat"
    LLAMA_2_70B_CHAT = "llama-2-70b-chat"
    MISTRAL_7B_INSTRUCT_V0_2 = "mistral-7b-instruct-v0-2"


class CompletionsModel(str, Enum):
    """Model choices for completions."""

    LLAMA_2_13B = "llama-2-13b"
    LLAMA_2_70B = "llama-2-70b"
    MISTRAL_7B_V0_1 = "mistral-7b-v0-1"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli GPTJ Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Dict, List, cast

import torch
from transformers import GPTJConfig  # type: ignore[import]

from friendli.errors import CheckpointConversionError, NotSupportedCheckpointError
from friendli.logging import logger
from friendli.modules.converter.base import (
    DECODER_PREFIX,
    SUPPORTED_GELU_FAMILY,
    DecoderOnlyConverter,
    DecoderOnlyLoraConverter,
)
from friendli.modules.converter.interface import RotaryEmbeddingConversionInterface
from friendli.modules.converter.schema import ConvertInfo


class GPTJForCausalLMLoraConverter(DecoderOnlyLoraConverter):
    """GPTJForCausalLM LoRA Converter Class."""

    @property
    def adapter_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for LoRA adapter modules in GPTJ."""
        convert_info_list = []
        for i in range(self.converter.decoder_layer_num):
            layer_prefix = f"{self.converter.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/lora/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}attn.q_proj.lora_A.default.weight"
                        ],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}query_A/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}attn.q_proj.lora_B.default.weight"
                        ],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}query_B/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}attn.v_proj.lora_A.default.weight"
                        ],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}value_A/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}attn.v_proj.lora_B.default.weight"
                        ],
                        data_type=self.converter.data_type,
                        converted_name=f"{converted_prefix}value_B/weight:0",
                        reshape_fn=self.lora_weight_reshape,
                    ),
                ]
            )
        return convert_info_list


class GPTJForCausalLMConverter(
    DecoderOnlyConverter, RotaryEmbeddingConversionInterface
):
    """GPTJForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if GPTJ architectures' config can be converted to Friendli format."""
        super().check_config()
        try:
            if (
                cast(GPTJConfig, self.config).activation_function
                not in SUPPORTED_GELU_FAMILY
            ):
                raise NotSupportedCheckpointError(
                    invalid_option="'activation_function="
                    f"{cast(GPTJConfig, self.config).activation_function}'",
                    valid_options=SUPPORTED_GELU_FAMILY,
                )
            if cast(GPTJConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if cast(GPTJConfig, self.config).layer_norm_epsilon != 1e-5:
                raise NotSupportedCheckpointError(
                    invalid_option="'layer_norm_epsilon="
                    f"{cast(GPTJConfig, self.config).layer_norm_epsilon}'",
                    valid_options=[1e-5],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_reshape(
        self,
        params: List[torch.Tensor],
    ) -> torch.Tensor:
        """qkv_weight_reshape for GPTJ's attention layer."""
        assert len(params) == 3
        qkv_weight = torch.cat(
            params,
            dim=0,
        )
        qkv_weight = qkv_weight.transpose(0, 1)
        return qkv_weight

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(GPTJConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The GPTJ model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.n_positions,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": config.n_positions,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "rope_theta": self.rotary_emb_base,
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "gpt-j"

    @property
    def non_transformer_convert_info_list(
        self,
    ) -> List[ConvertInfo]:
        """The list of conversion informations for non-transformer blocks in GPTJ."""
        return [
            ConvertInfo(
                param_names=["transformer.wte.weight"],
                data_type=self.data_type,
                converted_name="wte/weight:0",
                reshape_fn=self.token_embed_weight_reshape,
            ),
            ConvertInfo(
                param_names=["transformer.ln_f.weight"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/gamma:0",
                reshape_fn=self.ln_weight_reshape,
            ),
            ConvertInfo(
                param_names=["transformer.ln_f.bias"],
                data_type=self.data_type,
                converted_name=f"{DECODER_PREFIX}/ln_f/beta:0",
                reshape_fn=self.ln_bias_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.weight"],
                data_type=self.data_type,
                converted_name="head_fc/weight:0",
                reshape_fn=self.head_weight_reshape,
            ),
            ConvertInfo(
                param_names=["lm_head.bias"],
                data_type=self.data_type,
                converted_name="head_fc/bias:0",
                reshape_fn=self.head_weight_reshape,
            ),
        ]

    @property
    def decoder_convert_info_list(self) -> List[ConvertInfo]:
        """The list of conversion informations for transformer modules in GPTJ."""
        convert_info_list = []
        for i in range(self.decoder_layer_num):
            layer_prefix = f"{self.decoder_layer_prefix}{i}."
            converted_prefix = f"{DECODER_PREFIX}/h_._{i}/"
            convert_info_list.extend(
                [
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ln_1.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/gamma:0",
                        reshape_fn=self.ln_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}ln_1.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}ln_1/beta:0",
                        reshape_fn=self.ln_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc_in.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc_out.bias"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/bias:0",
                        reshape_fn=self.linear_bias_reshape,
                    ),
                    ConvertInfo(
                        param_names=[
                            f"{layer_prefix}attn.q_proj.weight",
                            f"{layer_prefix}attn.k_proj.weight",
                            f"{layer_prefix}attn.v_proj.weight",
                        ],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_attn/weight:0",
                        reshape_fn=self.qkv_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}attn.out_proj.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}attn/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc_in.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_fc/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                    ConvertInfo(
                        param_names=[f"{layer_prefix}mlp.fc_out.weight"],
                        data_type=self.data_type,
                        converted_name=f"{converted_prefix}mlp/c_proj/weight:0",
                        reshape_fn=self.linear_weight_reshape,
                    ),
                ]
            )

        return convert_info_list

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before GPTJ's transformer module number."""
        return "transformer.h."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in GPTJ."""
        return cast(GPTJConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in GPTJ."""
        return cast(GPTJConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in GPTJ."""
        return cast(GPTJConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in gpt-j."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head siez of GPTJ."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def decoder_ff_intermediate_size(self) -> int:
        """The intermediate size of the linear layer in codegen MLP."""
        return self.decoder_hidden_size * 4

    @property
    def rotary_dim(self) -> int:
        """The rotary dim in GPTJ."""
        return cast(GPTJConfig, self.config).rotary_dim

    @property
    def rotary_emb_base(self) -> float:
        """The rotary emb base in GPTJ."""
        return 10000.0



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Cloud Services."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import boto3
from azure.storage.blob import BlobServiceClient
from botocore.exceptions import ClientError
from mypy_boto3_s3.client import S3Client
from typing_extensions import TypeAlias

from friendli.enums import CloudType, StorageType
from friendli.utils.format import secho_error_and_exit

_CloudClient: TypeAlias = Union[S3Client, BlobServiceClient]
T = TypeVar("T", bound=_CloudClient)


storage_type_map: Dict[StorageType, str] = {
    StorageType.S3: "aws",
    StorageType.BLOB: "azure.blob",
    StorageType.GCS: "gcp",
    StorageType.FAI: "fai",
}


storage_type_map_inv: Dict[str, StorageType] = {
    "aws": StorageType.S3,
    "azure.blob": StorageType.BLOB,
    "gcp": StorageType.GCS,
    "fai": StorageType.FAI,
}


GCP_REGION_NAMES = [
    "asia-east1-a",
    "asia-east1-b",
    "asia-east1-c",
    "asia-east2-a",
    "asia-east2-b",
    "asia-east2-c",
    "asia-northeast1-a",
    "asia-northeast1-b",
    "asia-northeast1-c",
    "asia-northeast2-a",
    "asia-northeast2-b",
    "asia-northeast2-c",
    "asia-northeast3-a",
    "asia-northeast3-b",
    "asia-northeast3-c",
    "asia-south1-a",
    "asia-south1-b",
    "asia-south1-c",
    "asia-south2-a",
    "asia-south2-b",
    "asia-south2-c",
    "asia-southeast1-a",
    "asia-southeast1-b",
    "asia-southeast1-c",
    "asia-southeast2-a",
    "asia-southeast2-b",
    "asia-southeast2-c",
    "australia-southeast1-a",
    "australia-southeast1-b",
    "australia-southeast1-c",
    "australia-southeast2-a",
    "australia-southeast2-b",
    "australia-southeast2-c",
    "europe-central2-a",
    "europe-central2-b",
    "europe-central2-c",
    "europe-north1-a",
    "europe-north1-b",
    "europe-north1-c",
    "europe-west1-b",
    "europe-west1-c",
    "europe-west1-d",
    "europe-west2-a",
    "europe-west2-b",
    "europe-west2-c",
    "europe-west3-a",
    "europe-west3-b",
    "europe-west3-c",
    "europe-west4-a",
    "europe-west4-b",
    "europe-west4-c",
    "europe-west6-a",
    "europe-west6-b",
    "europe-west6-c",
    "northamerica-northeast1-a",
    "northamerica-northeast1-b",
    "northamerica-northeast1-c",
    "northamerica-northeast2-a",
    "northamerica-northeast2-b",
    "northamerica-northeast2-c",
    "southamerica-east1-a",
    "southamerica-east1-b",
    "southamerica-east1-c",
    "southamerica-west1-a",
    "southamerica-west1-b",
    "southamerica-west1-c",
    "us-central1-a",
    "us-central1-b",
    "us-central1-c",
    "us-central1-f",
    "us-east1-b",
    "us-east1-c",
    "us-east1-d",
    "us-east4-a",
    "us-east4-b",
    "us-east4-c",
    "us-west1-a",
    "us-west1-b",
    "us-west1-c",
    "us-west2-a",
    "us-west2-b",
    "us-west2-c",
    "us-west3-a",
    "us-west3-b",
    "us-west3-c",
    "us-west4-a",
    "us-west4-b",
]

AZURE_REGION_NAMES = [
    "eastus",
    "eastus2",
    "southcentralus",
    "westus2",
    "westus3",
    "australiaeast",
    "southeastasia",
    "northeurope",
    "swedencentral",
    "uksouth",
    "westeurope",
    "centralus",
    "northcentralus",
    "westus",
    "southafricanorth",
    "centralindia",
    "eastasia",
    "japaneast",
    "jioindiawest",
    "koreacentral",
    "canadacentral",
    "francecentral",
    "germanywestcentral",
    "norwayeast",
    "switzerlandnorth",
    "uaenorth",
    "brazilsouth",
    "centralusstage",
]


AWS_REGION_NAMES = [
    "us-east-2",
    "us-east-1",
    "us-west-1",
    "us-west-2",
    "af-south-1",
    "ap-east-1",
    "ap-south-2",
    "ap-southeast-3",
    "ap-southeast-4",
    "ap-south-1",
    "ap-northeast-3",
    "ap-northeast-2",
    "ap-southeast-1",
    "ap-southeast-2",
    "ap-northeast-1",
    "ca-central-1",
    "eu-central-1",
    "eu-west-1",
    "eu-west-2",
    "eu-south-1",
    "eu-west-3",
    "eu-south-1",
    "eu-north-1",
    "eu-central-2",
    "me-south-1",
    "me-central-1",
    "sa-east-1",
    "us-gov-east-1",
    "us-gov-west-1",
]


FAI_REGION_NAMES = [
    "",
]


storage_region_map = {
    StorageType.S3: AWS_REGION_NAMES,
    StorageType.BLOB: AZURE_REGION_NAMES,
    StorageType.GCS: GCP_REGION_NAMES,
    StorageType.FAI: FAI_REGION_NAMES,
}


cloud_region_map = {
    CloudType.AWS: AWS_REGION_NAMES,
    CloudType.AZURE: AZURE_REGION_NAMES,
    CloudType.GCP: GCP_REGION_NAMES,
}


class CloudStorageClient(ABC, Generic[T]):
    """Cloud storage client interface."""

    def __init__(self, client: T) -> None:
        """Initialize cloud storage client."""
        self.client = client

    @abstractmethod
    def list_storage_files(
        self, storage_name: str, path_prefix: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """List all file objects in the storage.

        Args:
            storage_name (str): Storage name
            path_prefix (Optional[str], optional): Direcotry path under the storage.

        Returns:
            List[Dict[str, Any]]: A list of object info.

        """


class AWSCloudStorageClient(CloudStorageClient[S3Client]):
    """AWS S3 client."""

    def _check_aws_bucket_exists(self, storage_name: str) -> bool:
        try:
            self.client.head_bucket(Bucket=storage_name)
            return True
        except ClientError:
            # include both Forbidden access, Not Exists
            return False

    def list_storage_files(
        self, storage_name: str, path_prefix: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """List objects in the storage."""
        if not self._check_aws_bucket_exists(storage_name):
            secho_error_and_exit(f"Bucket {storage_name} does not exist")

        file_list = []
        if path_prefix is None:
            resp = self.client.list_objects(Bucket=storage_name)
        else:
            resp = self.client.list_objects(Bucket=storage_name, Prefix=path_prefix)

        if "Contents" not in resp:
            secho_error_and_exit(
                f"No file exists at {path_prefix} in the bucket({storage_name})"
            )
        object_contents = resp["Contents"]
        for object_content in object_contents:
            try:
                object_key = object_content["Key"]
                name = object_key.split("/")[-1]
                if not name:
                    continue  # skip directory
                file_list.append(
                    {
                        "name": name,
                        "path": object_key,
                        "mtime": object_content["LastModified"].isoformat(),
                        "size": object_content["Size"],
                    }
                )
            except KeyError:
                secho_error_and_exit("Unexpected S3 error")

        if not file_list:
            secho_error_and_exit(f"No file exists in Bucket {storage_name}")

        return file_list


class AzureCloudStorageClient(CloudStorageClient[BlobServiceClient]):
    """Azure blob storage client."""

    def list_storage_files(
        self, storage_name: str, path_prefix: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """List objects in the storage."""
        container_client = self.client.get_container_client(storage_name)
        if not container_client.exists():
            secho_error_and_exit(f"Container {storage_name} does not exist")

        file_list = []
        prefix_option = (
            {"name_starts_with": path_prefix} if path_prefix is not None else {}
        )
        object_contents = container_client.list_blobs(**prefix_option)
        for object_content in object_contents:
            object_name = object_content["name"]
            name = object_name.split("/")[-1]
            if not name:
                continue  # skip directory
            file_list.append(
                {
                    "name": name,
                    "path": object_name,
                    "mtime": object_content["last_modified"].isoformat(),
                    "size": object_content["size"],
                }
            )

        if not file_list:
            secho_error_and_exit(f"No file exists in Bucket {storage_name}")

        return file_list


def build_s3_client(credential_json: Dict[str, str]) -> S3Client:
    """Build AWS S3 client."""
    return boto3.client(
        "s3",
        aws_access_key_id=credential_json["aws_access_key_id"],
        aws_secret_access_key=credential_json["aws_secret_access_key"],
        region_name=credential_json.get("aws_default_region", None),
    )


def build_blob_client(credential_json: Dict[str, str]) -> BlobServiceClient:
    """Build Azure blob storage client."""
    url = f"https://{credential_json['storage_account_name']}.blob.core.windows.net/"
    return BlobServiceClient(
        account_url=url, credential=credential_json["storage_account_key"]
    )


# TODO: Add GCP support
vendor_client_map: Dict[
    StorageType,
    Tuple[Type[CloudStorageClient], Callable[[Dict[str, str]], _CloudClient]],
] = {
    StorageType.S3: (AWSCloudStorageClient, build_s3_client),
    StorageType.BLOB: (AzureCloudStorageClient, build_blob_client),
}


def build_storage_client(
    vendor: StorageType, credential_json: Dict[str, Any]
) -> CloudStorageClient:
    """Build a cloud storage client."""
    cls, client_build_fn = vendor_client_map[vendor]
    client = client_build_fn(credential_json)
    return cls(client)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Credential CLI."""

# pylint: disable=line-too-long

from __future__ import annotations

import json
import sys
from uuid import UUID

import typer

from friendli.configurator.credential import CredentialInteractiveConfigurator
from friendli.enums import CredType
from friendli.formatter import PanelFormatter, TableFormatter
from friendli.sdk.resource.credential import Credential
from friendli.utils.decorator import check_api
from friendli.utils.format import secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)
create_app = typer.Typer(
    help="Create credential",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

app.add_typer(create_app, name="register")

table_formatter = TableFormatter(
    name="Credentials",
    fields=["id", "name", "type", "created_at", "owner_type"],
    headers=["ID", "Name", "Type", "Created At", "Scope"],
)

panel_formatter = PanelFormatter(
    name="Overview",
    fields=["id", "name", "type", "created_at", "owner_type"],
    headers=["ID", "Name", "Type", "Created At", "Scope"],
)


S3_DOC_LINK = (
    "https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-quickstart.html"
)
AZURE_BLOB_DOC_LINK = "https://docs.microsoft.com/en-us/azure/storage/common/storage-account-keys-manage?tabs=azure-portal"  # pylint: disable=line-too-long
GCP_DOC_LINK = (
    "https://cloud.google.com/iam/docs/creating-managing-service-account-keys"
)


@create_app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    interactive: bool = typer.Option(
        False, "--interactive", help="Use interactive mode."
    ),
):
    """Common callback for the interactive option."""
    if not interactive:
        if ctx.invoked_subcommand is None:
            secho_error_and_exit(
                f"You should provide one of the credential types {[ e.value for e in CredType ]}"
                " as an argument, or use '--interactive' option. "
                "Run 'friendli credential --help' for more details."
            )
        return

    configurator = CredentialInteractiveConfigurator()

    configurator.start_interaction()
    name, cred_type, value = configurator.render()

    client = Credential()
    cred = client.create(cred_type=cred_type, name=name, value=value)

    typer.secho("Credential created successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([cred.model_dump()])
    sys.exit(0)


@create_app.command()
@check_api
def s3(
    name: str = typer.Option(..., "--name", "-n", help="Name of the new credential."),
    aws_access_key_id: str = typer.Option(
        ..., help=f"[AWS] AWS Acess Key ID. Please refer to {S3_DOC_LINK}."
    ),
    aws_secret_access_key: str = typer.Option(
        ..., help=f"[AWS] AWS Secret Access Key. Please refer to {S3_DOC_LINK}."
    ),
    aws_default_region: str = typer.Option(
        ..., help=f"[AWS] Default region name. Please refer to {S3_DOC_LINK}."
    ),
):
    """Creates a credential for AWS S3.

    To use checkpoints in the S3 buckets of your own AWS account, you must provide
    the credential to access the S3 buckets.

    :::tip
    How to get the value of AWS access key ID, AWS secret access key, and AWS default
    region:

    1. Go to the AWS console website.
    2. Sign in with your AWS account with the S3 bucket you want to use.
    3. Go to the [security credential page](https://console.aws.amazon.com/iam/home?#security_credential) and click **Create access key** at the **AWS IAM credentials** tab.
    4. Get a new access key ID and secret access key.
    5. For the AWS default region, use the AWS region code in the [region code table](https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/using-regions-availability-zones.html#concepts-regions).
    :::

    :::info
    You should have a proper IAM permission to read objects in a S3 bucket that you want to use in Friendli. Please refer to the [AWS IAM user guide](https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_examples_s3_rw-bucket.html) to learn how to set the IAM policy for S3.
    :::

    """
    value = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "aws_default_region": aws_default_region,
    }
    client = Credential()
    cred = client.create(cred_type=CredType.S3, name=name, value=value)
    panel_formatter.render([cred.model_dump()])


@create_app.command()
@check_api
def azure_blob(
    name: str = typer.Option(..., "--name", "-n", help="Name of the new credential."),
    storage_account_name: str = typer.Option(
        ..., help="[Azure] Azure Blob storage account name."
    ),
    storage_account_key: str = typer.Option(
        ...,
        help=f"[Azure] Azure Blob storage account access key. Please refer to {AZURE_BLOB_DOC_LINK}.",
    ),
):
    """Creates a credential for Azure Blob storage.

    To use checkpoints in the Azure Blob Storage container, you must provide the
    credential to access the storage account.

    :::tip
    To get the value of storage account key:

    1. Go to Azure portal website.
    2. Sign in your Azure account with the storage you want to use.
    3. Go to the **Storage account** page and click the storage account you want.
    4. At the side navigation tab, click **Access keys**.
    5. Copy the **Key** value for any key between **key1** and **key2**.
    6. Use that value for the `--storage-acount-key` option.
    :::

    :::info
    Azure Blob Storage has three types of resources (in a hierarchical order):

    - The storage account
    - A container in the storage account
    - A blob in a container

    If you create an `azure-blob` credential, you can use the credential for all
    containers in the storage account. Please refer to the [Azure Blob Storage documentation](https://docs.microsoft.com/azure/storage/blobs/storage-blobs-introduction#blob-storage-resources)
    for more information.
    :::

    """
    value = {
        "storage_account_name": storage_account_name,
        "storage_account_key": storage_account_key,
    }
    client = Credential()
    cred = client.create(cred_type=CredType.BLOB, name=name, value=value)
    panel_formatter.render([cred.model_dump()])


@create_app.command()
@check_api
def gcs(
    name: str = typer.Option(..., "--name", "-n", help="Name of the new credential."),
    service_account_key_file: typer.FileText = typer.Option(
        ...,
        help=f"[GCP] Path to GCP Service Account Key JSON file. Please refer to {GCP_DOC_LINK}.",
    ),
):
    r"""Creates a credential for Google Cloud Storage.

    To use checkpoints in the Google Cloud Storage, you must create a
    credential to access the GCS buckets.


    :::tip
    How to get a JSON file containing your GCP service account key for the `--service-account-key-file` option:

    1. Go to Google Cloud Platform Console, and sign in.
    2. Go to the **Service accounts** page.
    3. Select the project with the bucket you want to use.
    4. On the **Service accounts** page, click the email address of the service account whose keys you want to list.
    5. Click **Keys**. The console should display a list of keys for your service account, including metadata for each key.
    6. Click the **Add key** dropdown menu and select **Create new key**.
    7. For **Key type**, select "JSON" and click **Create**, which downloads a JSON file with the following format. Use the path to that file for the `--service-account-key-file` option.

    ```json
    {
        "type": "service_account",
        "project_id": "project-id",
        "private_key_id": "key-id",
        "private_key": "-----BEGIN PRIVATE KEY-----\nprivate-key\n-----END PRIVATE KEY-----\n",
        "client_email": "service-account-email",
        "client_id": "client-id",
        "auth_uri": "https://accounts.google.com/o/oauth2/auth",
        "token_uri": "https://accounts.google.com/o/oauth2/token",
        "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
        "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/service-account-email"
    }
    ```
    :::

    :::info
    You should have a proper IAM role to read objects in a bucket you want to use in Friendli. Please see the [GCS IAM role documentation](https://cloud.google.com/storage/docs/access-control/iam-roles).
    :::

    """
    try:
        value = json.load(service_account_key_file)
    except json.JSONDecodeError as exc:
        secho_error_and_exit(f"Error occurred while parsing JSON file... {exc}")
    del value["type"]
    client = Credential()
    cred = client.create(cred_type=CredType.GCS, name=name, value=value)
    panel_formatter.render([cred.model_dump()])


# pylint: disable=redefined-builtin
@app.command()
@check_api
def list(
    cred_type: CredType = typer.Option(
        None, "--cred-type", "-t", help="Filter credentials by the type."
    ),
):
    """Lists credentials."""
    client = Credential()
    creds = client.list(cred_type=cred_type)
    table_formatter.render([cred.model_dump() for cred in creds])


@app.command()
@check_api
def update(
    credential_id: UUID = typer.Argument(..., help="ID of credential to update.")
):
    """Updates data of a credential.

    Only interactive mode is supported for now.

    """
    configurator = CredentialInteractiveConfigurator()
    configurator.start_interaction_for_update(credential_id)
    name, _, value = configurator.render()

    client = Credential()
    cred = client.edit(id=credential_id, name=name, value=value)
    typer.secho("Credential updated successfully!", fg=typer.colors.BLUE)
    panel_formatter.render([cred.model_dump()])


@app.command()
@check_api
def delete(
    credential_id: UUID = typer.Argument(..., help="ID of credential to delete."),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Forcefully deletes credential without confirmation prompt.",
    ),
):
    """Deletes a credential."""
    if not force:
        do_delete = typer.confirm("Are you sure to delete credential?")
        if not do_delete:
            raise typer.Abort()

    client = Credential()
    client.delete(id=credential_id)

    typer.secho(
        f"Credential ({credential_id}) is deleted successfully!", fg=typer.colors.BLUE
    )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Deployment Schemas."""

from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from friendli.enums import DeploymentReplicaStatus, DeploymentStatus, DeploymentType


class V1VirtualMachine(BaseModel):
    """V1 virtual machine schema."""

    name: Optional[str]
    gpu_type: Optional[str]
    gpu_memory: float
    allocated_gpus: int
    vcpu: int
    cpu_memory: float


class InferenceServer(BaseModel):
    """V1 inference server schema."""

    name: str
    repo: str
    tag: str


class V1DeploymentConfig(BaseModel):
    """V1 deployment config schema."""

    model_config = ConfigDict(protected_namespaces=())

    name: str = Field(..., max_length=127)
    description: Optional[str] = Field(None, max_length=127)
    model_id: UUID
    deployment_type: DeploymentType
    vm: V1VirtualMachine
    project_id: UUID
    inference_server_type: InferenceServer
    orca_config: Optional[Dict[str, Any]]
    triton_config: Optional[Dict[str, Any]]
    proxy_rate_limit: int = Field(0, ge=0)
    download_ckpt: bool
    total_gpus: int = Field(1, ge=0)
    ckpt_config: Dict[str, Any]
    cloud: str
    region: str
    cluster_name: Optional[str] = Field(None, max_length=127)
    scaler_config: Dict[str, Any]
    infrequest_perm_check: bool
    infrequest_log: bool


class V1Deployment(BaseModel):
    """V1 deployment schema."""

    deployment_id: str
    version: int
    update_msg: str
    description: Optional[str]
    namespace: str
    user_id: UUID
    project_id: UUID
    config: V1DeploymentConfig
    type: DeploymentType
    ready_replicas: int
    replica_status: List[DeploymentReplicaStatus]
    status: DeploymentStatus
    vms: List[V1VirtualMachine]
    start: datetime
    end: Optional[datetime]
    endpoint: str  # TODO: Use HttpUrl
    error_msg: str
    updated_time: datetime



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli CLI."""

# pylint: disable=line-too-long

from __future__ import annotations

import requests
import typer
from requests import HTTPError, Response

from friendli.auth import TokenType, clear_tokens, get_token, update_token
from friendli.cli import api, checkpoint
from friendli.client.project import ProjectClient
from friendli.client.user import UserClient, UserGroupClient, UserMFAClient
from friendli.context import (
    get_current_project_id,
    project_context_path,
    set_current_group_id,
)
from friendli.di.injector import get_injector
from friendli.errors import AuthTokenNotFoundError
from friendli.formatter import PanelFormatter
from friendli.utils.decorator import check_api
from friendli.utils.format import secho_error_and_exit
from friendli.utils.request import DEFAULT_REQ_TIMEOUT
from friendli.utils.url import URLProvider
from friendli.utils.version import get_installed_version

app = typer.Typer(
    help="Supercharge Generative AI Serving 🚀",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
    pretty_exceptions_enable=False,
)

# Cloud-related features are temporarily disabled until account system is integrated with Friendli suite.
# app.add_typer(catalog.app, name="catalog", help="Manage catalog")
# app.add_typer(credential.app, name="credential", help="Manage credentials")
# app.add_typer(gpu.app, name="gpu", help="Manage GPUs")
# app.add_typer(deployment.app, name="deployment", help="Manage deployments")
# app.add_typer(project.app, name="project", help="Manage projects")
# app.add_typer(group.app, name="org", help="Manage organizations")
# app.add_typer(key.app, name="key", help="Manage api keys")
app.add_typer(checkpoint.app, name="checkpoint", help="Manage checkpoints")
app.add_typer(api.app, name="api", help="API call to endpoints")


user_panel_formatter = PanelFormatter(
    name="My Info",
    fields=["name", "email"],
    headers=["Name", "Email"],
)


# @app.command()
@check_api
def whoami():
    """Show my user info."""
    try:
        client = UserClient()
        info = client.get_current_user_info()
    except AuthTokenNotFoundError as exc:
        secho_error_and_exit(str(exc))

    user_panel_formatter.render([info])


# @app.command()
@check_api
def login(
    email: str = typer.Option(..., prompt="Enter your email"),
    password: str = typer.Option(..., prompt="Enter your password", hide_input=True),
):
    """Sign in."""
    injector = get_injector()
    url_provider = injector.get(URLProvider)
    r = requests.post(
        url_provider.get_web_backend_uri("/api/auth/cli/access_token"),
        json={"username": email, "password": password},
        timeout=DEFAULT_REQ_TIMEOUT,
    )
    try:
        resp = r.json()
    except requests.exceptions.JSONDecodeError:
        if r.status_code != 200:
            secho_error_and_exit(r.content.decode())
        secho_error_and_exit("Invalid response format.")

    if "code" in resp and resp["code"] == "mfa_required":
        mfa_token = resp["mfaToken"]
        client = UserMFAClient()
        # TODO: MFA type currently defaults to totp, need changes when new options are added
        client.initiate_mfa(mfa_type="totp", mfa_token=mfa_token)
        update_token(token_type=TokenType.MFA, token=mfa_token)
        typer.run(_mfa_verify)
    else:
        _handle_login_response(r, False)

    # Save user's organiztion context
    project_client = ProjectClient()
    user_group_client = UserGroupClient()

    try:
        org = user_group_client.get_group_info()
    except IndexError:
        secho_error_and_exit("You are not included in any organization.")
    org_id = org["id"]

    project_id = get_current_project_id()
    if project_id is not None:
        if project_client.check_project_membership(pf_project_id=project_id):
            project_org_id = project_client.get_project(pf_project_id=project_id)[
                "pf_group_id"
            ]
            if project_org_id != org_id:
                project_context_path.unlink(missing_ok=True)
        else:
            project_context_path.unlink(missing_ok=True)
    set_current_group_id(org_id)


# @app.command()
def logout():
    """Sign out."""
    clear_tokens()
    typer.secho("Successfully signed out.", fg=typer.colors.BLUE)


# @app.command()
@check_api
def passwd(
    old_password: str = typer.Option(
        ..., prompt="Enter your current password", hide_input=True
    ),
    new_password: str = typer.Option(
        ..., prompt="Enter your new password", hide_input=True
    ),
    confirm_password: str = typer.Option(
        ..., prompt="Enter the new password again (confirmation)", hide_input=True
    ),
):
    """Change password."""
    if old_password == new_password:
        secho_error_and_exit("The current password is the same with the new password.")
    if new_password != confirm_password:
        secho_error_and_exit("Passwords did not match.")
    try:
        client = UserClient()
        client.change_password(old_password, new_password)
    except AuthTokenNotFoundError as exc:
        secho_error_and_exit(str(exc))

    typer.secho("Password is changed successfully!", fg=typer.colors.BLUE)


@app.command()
def version():
    """Check the installed package version."""
    installed_version = get_installed_version()
    typer.echo(installed_version)


def _mfa_verify(_, code: str = typer.Option(..., prompt="Enter MFA Code")):
    injector = get_injector()
    url_provider = injector.get(URLProvider)

    mfa_token = get_token(TokenType.MFA)
    # TODO: MFA type currently defaults to totp, need changes when new options are added
    mfa_type = "totp"
    username = f"mfa://{mfa_type}/{mfa_token}"
    r = requests.post(
        url_provider.get_web_backend_uri("/api/auth/cli/access_token"),
        json={"username": username, "password": code},
        timeout=DEFAULT_REQ_TIMEOUT,
    )
    _handle_login_response(r, True)


def _handle_login_response(r: Response, mfa: bool):
    try:
        r.raise_for_status()
        update_token(token_type=TokenType.ACCESS, token=r.json()["accessToken"])
        update_token(token_type=TokenType.REFRESH, token=r.json()["refreshToken"])

        typer.echo("\n\nLogin success!")
        typer.echo("Welcome back to...")
        typography = r"""
 _____                         _  _
|  ___|_  _(_) ___  _  __    _| || |(_)  
| |__ | '__| |/ _ \| '__  \/ _  || || |
|  __|| |  | |  __/| |  | | (_) || || |
|_|   |_|  |_|\___||_|  |_|\___/ |_||_|
"""
        typer.secho(typography, fg=typer.colors.BLUE)
    except HTTPError:
        if mfa:
            secho_error_and_exit("Login failed... Invalid MFA Code.")
        else:
            secho_error_and_exit(
                "Login failed... Please check your email and password."
            )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=redefined-builtin, too-many-locals, too-many-arguments, line-too-long

"""Friendli Checkpoint CLI."""

from __future__ import annotations

import os
import random
import string
from typing import List, Optional, cast
from uuid import UUID

import typer
import yaml

from friendli.enums import (
    CatalogImportMethod,
    CheckpointCategory,
    CheckpointDataType,
    CheckpointFileType,
    StorageType,
)
from friendli.errors import (
    CheckpointConversionError,
    InvalidAttributesError,
    InvalidConfigError,
    InvalidPathError,
    NotFoundError,
    NotSupportedError,
)
from friendli.formatter import (
    JSONFormatter,
    PanelFormatter,
    TableFormatter,
    TreeFormatter,
)
from friendli.sdk.resource.catalog import Catalog
from friendli.sdk.resource.checkpoint import Checkpoint
from friendli.utils.decorator import check_api
from friendli.utils.format import (
    datetime_to_pretty_str,
    get_translated_checkpoint_status,
    secho_error_and_exit,
)

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

table_formatter = TableFormatter(
    name="Checkpoints",
    fields=[
        "id",
        "name",
        "model_category",
        "iteration",
        "created_at",
    ],
    headers=[
        "ID",
        "Name",
        "Source",
        "Iteration",
        "Created At",
    ],
)
panel_formatter = PanelFormatter(
    name="Overview",
    fields=[
        "id",
        "name",
        "model_category",
        "forms[0].vendor",
        "forms[0].storage_name",
        "iteration",
        "forms[0].form_category",
        "created_at",
        "status",
    ],
    headers=[
        "ID",
        "Name",
        "Source",
        "Cloud",
        "Storage Name",
        "Iteration",
        "Format",
        "Created At",
        "Status",
    ],
)
json_formatter = JSONFormatter(name="Attributes")
tree_formatter = TreeFormatter(name="Files")


# @app.command()
@check_api
def list(
    source: Optional[CheckpointCategory] = typer.Option(
        None,
        "--source",
        "-s",
        help="Source of checkpoints.",
    ),
    limit: int = typer.Option(
        20,
        "--limit",
        "-l",
        help="The number of recent checkpoints to see.",
    ),
    deleted: bool = typer.Option(
        False, "--deleted", "-d", help="Shows deleted checkpoint."
    ),
):
    """Lists all checkpoints that belong to a user's organization."""
    client = Checkpoint()
    checkpoints = client.list(category=source, limit=limit, deleted=deleted)
    ckpt_dicts = []
    for ckpt in checkpoints:
        ckpt_dict = ckpt.model_dump()
        ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
        ckpt_dicts.append(ckpt_dict)

    table_formatter.render(ckpt_dicts)


# @app.command()
@check_api
def view(
    checkpoint_id: UUID = typer.Argument(
        ..., help="ID of checkpoint to inspect in detail."
    )
):
    """Shows details of a checkpoint."""
    client = Checkpoint()
    ckpt = client.get(id=checkpoint_id)
    ckpt_dict = ckpt.model_dump()

    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    ckpt_dict["status"] = get_translated_checkpoint_status(ckpt)

    panel_formatter.render([ckpt_dict])
    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


# @app.command()
@check_api
def create(
    name: str = typer.Option(
        ..., "--name", "-n", help="Name of your checkpoint to create."
    ),
    credential_id: UUID = typer.Option(
        ...,
        "--credential-id",
        "-i",
        help="ID of crendential to access cloud storage.",
    ),
    cloud_storage: StorageType = typer.Option(
        ...,
        "--cloud-storage",
        "-c",
        help="The cloud storage vendor where the checkpoint is uploaded.",
    ),
    storage_name: str = typer.Option(
        ...,
        "--storage-name",
        "-s",
        help="The name of the cloud storage where the checkpoint is uploaded.",
    ),
    region: str = typer.Option(
        ...,
        "--region",
        "-r",
        help="The cloud storage region where the checkpoint is uploaded.",
    ),
    storage_path: Optional[str] = typer.Option(
        None,
        "--storage-path",
        "-p",
        help="File or directory path to cloud storage. Defaults to the root of the storage.",
    ),
    iteration: Optional[int] = typer.Option(
        None, "--iteration", help="The iteration number of the checkpoint."
    ),
    attr_file: Optional[str] = typer.Option(
        None,
        "--attr-file",
        "-f",
        help=(
            "Path to the file containing checkpoint attributes. The file should be in "
            "YAML format."
        ),
    ),
):
    """Creates a checkpoint by registering checkpoint files in user's cloud storage to Friendli.

    An example of attribute files is as follows.
    You have the flexibility to modify the value of each field according to your preferences,
    but it is mandatory to fill in every field.

    ```yaml
    # blenderbot
    model_type: blenderbot
    dtype: fp16
    head_size: 80
    num_heads: 32
    hidden_size: 2560
    ff_intermediate_size: 10240
    num_encoder_layers: 2
    num_decoder_layers: 24
    max_input_length: 128
    max_output_length: 128
    vocab_size: 8008
    eos_token: 2
    decoder_start_token: 1

    # bloom
    model_type: bloom
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 30
    max_length: 2048
    vocab_size: 250880
    eos_token: 2

    # gpt
    model_type: gpt
    dtype: fp16
    head_size: 64
    num_heads: 25
    num_layers: 48
    max_length: 1024
    vocab_size: 50257
    eos_token: 50256

    # gpt-j
    model_type: gpt-j
    dtype: fp16
    head_size: 256
    rotary_dim: 64
    num_heads: 16
    num_layers: 28
    max_length: 2048
    vocab_size: 50400
    eos_token: 50256

    # gpt-neox
    model_type: gpt-neox
    dtype: fp16
    head_size: 128
    rotary_dim: 32
    num_heads: 40
    num_layers: 36
    max_length: 2048
    vocab_size: 50280
    eos_token: 0

    # llama
    model_type: llama
    dtype: fp16
    head_size: 128
    rotary_dim: 128
    num_heads: 32
    num_kv_heads: 32
    num_layers: 32
    ff_intermediate_size: 11008
    max_length: 2048
    vocab_size: 32000
    eos_token: 1

    # opt
    model_type: opt
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 32
    max_length: 2048
    vocab_size: 50272
    eos_token: 2

    # t5
    model_type: t5
    dtype: fp16
    head_size: 128
    num_heads: 32
    hidden_size: 1024
    ff_intermediate_size: 16384
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32100
    eos_token: 1
    decoder_start_token: 0

    # t5-v1_1
    model_type: t5-v1_1
    dtype: fp16
    head_size: 64
    num_heads: 32
    hidden_size: 2048
    ff_intermediate_size: 5120
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32128
    eos_token: 1
    decoder_start_token: 0
    ```

    :::tip
    If you want to use tokenizer, `tokenizer.json` file should be in the same directory
    as checkpoint file(`*.h5`).
    :::

    """
    client = Checkpoint()
    try:
        ckpt = client.create(
            name=name,
            credential_id=credential_id,
            cloud_storage=cloud_storage,
            region=region,
            storage_name=storage_name,
            storage_path=storage_path,
            iteration=iteration,
            attr_file_path=attr_file,
        )
    except (InvalidConfigError, NotSupportedError, InvalidAttributesError) as exc:
        secho_error_and_exit(str(exc))

    ckpt_dict = ckpt.model_dump()
    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    ckpt_dict["status"] = get_translated_checkpoint_status(ckpt)

    panel_formatter.render([ckpt_dict])
    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


# @app.command()
@check_api
def delete(
    checkpoint_ids: List[UUID] = typer.Argument(
        ...,
        help=(
            "IDs of checkpoint to delete. "
            "When multiple IDs are provided in a space-separated string format, all "
            "corresponding checkpoints will be deleted."
        ),
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Forcefully delete checkpoint without confirmation prompt.",
    ),
):
    """Deletes the existing checkpoint.

    :::info
    Deleting linked checkpoints created with `friendli checkpoint create` simply unlinks the
    checkpoint from the Friendli system. That is, it does not physically delete the
    checkpoint objects from the source cloud storage (your cloud storage).

    Uploaded checkpoints are physically deleted.
    :::

    """
    targets_str = ""
    for checkpoint_id in checkpoint_ids:
        targets_str += str(checkpoint_id)
        targets_str += "\n"

    if not force:
        do_delete = typer.confirm(
            f"Following checkpoints will be deleted:\n\n{targets_str}\n"
            "Are your sure to delete these?"
        )
        if not do_delete:
            raise typer.Abort()

    client = Checkpoint()
    for checkpoint_id in checkpoint_ids:
        client.delete(id=checkpoint_id)

    typer.secho("Checkpoints are deleted successfully!", fg=typer.colors.BLUE)


# @app.command()
@check_api
def download(
    checkpoint_id: UUID = typer.Argument(..., help="ID of checkpoint to download."),
    save_directory: Optional[str] = typer.Option(
        None,
        "--output",
        "-o",
        help="Output path to the directory to save checkpoint files.",
    ),
):
    """Downloads checkpoint files to local storage.

    :::info
    You cannot download checkpoints linked from your cloud storage.
    :::

    """
    client = Checkpoint()
    try:
        client.download(id=checkpoint_id, save_dir=save_directory)
    except InvalidPathError as exc:
        secho_error_and_exit(str(exc))


# @app.command()
@check_api
def upload(
    name: str = typer.Option(
        ..., "--name", "-n", help="Name of the checkpoint to upload"
    ),
    source_path: str = typer.Option(
        ..., "--source-path", "-p", help="Path to source file or directory to upload"
    ),
    iteration: Optional[int] = typer.Option(
        None, "--iteration", help="The iteration number of the checkpoint."
    ),
    attr_file: Optional[str] = typer.Option(
        None,
        "--attr-file",
        "-f",
        help="Path to the file containing the checkpoint attributes. The file should "
        "be in YAML format.",
    ),
    max_workers: int = typer.Option(
        min(32, (os.cpu_count() or 1) + 4),  # default of ``ThreadPoolExecutor``
        "--max-workers",
        "-w",
        help="The number of threads to upload files.",
    ),
):
    """Creates a checkpoint by uploading local checkpoint files.

    An example of attribute files is as follows.
    You have the flexibility to modify the value of each field according to your preferences,
    but it is mandatory to fill in every field.

    ```yaml
    # blenderbot
    model_type: blenderbot
    dtype: fp16
    head_size: 80
    num_heads: 32
    hidden_size: 2560
    ff_intermediate_size: 10240
    num_encoder_layers: 2
    num_decoder_layers: 24
    max_input_length: 128
    max_output_length: 128
    vocab_size: 8008
    eos_token: 2
    decoder_start_token: 1

    # bloom
    model_type: bloom
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 30
    max_length: 2048
    vocab_size: 250880
    eos_token: 2

    # gpt
    model_type: gpt
    dtype: fp16
    head_size: 64
    num_heads: 25
    num_layers: 48
    max_length: 1024
    vocab_size: 50257
    eos_token: 50256

    # gpt-j
    model_type: gpt-j
    dtype: fp16
    head_size: 256
    rotary_dim: 64
    num_heads: 16
    num_layers: 28
    max_length: 2048
    vocab_size: 50400
    eos_token: 50256

    # gpt-neox
    model_type: gpt-neox
    dtype: fp16
    head_size: 128
    rotary_dim: 32
    num_heads: 40
    num_layers: 36
    max_length: 2048
    vocab_size: 50280
    eos_token: 0

    # llama
    model_type: llama
    dtype: fp16
    head_size: 128
    rotary_dim: 128
    num_heads: 32
    num_kv_heads: 32
    num_layers: 32
    ff_intermediate_size: 11008
    max_length: 2048
    vocab_size: 32000
    eos_token: 1

    # opt
    model_type: opt
    dtype: fp16
    head_size: 128
    num_heads: 32
    num_layers: 32
    max_length: 2048
    vocab_size: 50272
    eos_token: 2

    # t5
    model_type: t5
    dtype: fp16
    head_size: 128
    num_heads: 32
    hidden_size: 1024
    ff_intermediate_size: 16384
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32100
    eos_token: 1
    decoder_start_token: 0

    # t5-v1_1
    model_type: t5-v1_1
    dtype: fp16
    head_size: 64
    num_heads: 32
    hidden_size: 2048
    ff_intermediate_size: 5120
    num_encoder_layers: 24
    num_decoder_layers: 24
    max_input_length: 512
    max_output_length: 512
    num_pos_emb_buckets: 32
    max_pos_distance: 128
    vocab_size: 32128
    eos_token: 1
    decoder_start_token: 0
    ```

    :::tip
    If you want to use tokenizer, `tokenizer.json` file should be in the same directory
    as checkpoint file(`*.h5`).
    :::

    """
    client = Checkpoint()
    try:
        ckpt = client.upload(
            name=name,
            source_path=source_path,
            iteration=iteration,
            attr_file_path=attr_file,
            max_workers=max_workers,
        )
    except (InvalidPathError, InvalidConfigError, InvalidAttributesError) as exc:
        secho_error_and_exit(str(exc))

    ckpt_dict = ckpt.model_dump()
    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    ckpt_dict["status"] = get_translated_checkpoint_status(ckpt)

    panel_formatter.render([ckpt_dict])
    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


# @app.command()
@check_api
def restore(
    checkpoint_id: UUID = typer.Argument(..., help="ID of checkpoint to restore.")
):
    """Restores deleted checkpoint (available within 24 hours from the deletion).

    When a uploaded checkpoint is deleted, it is not physically deleted right away.
    Instead, it is soft-deleted and can be restored within 24 hours after the deletion.
    After the 24-hour retention period, it is hard-deleted and cannot be restored.

    :::caution
    When you delete a linked checkpoint, you cannot restore it. Instead, relink the
    original checkpoint using `friendli checkpoint create`.
    :::

    """
    client = Checkpoint()
    try:
        client.restore(id=checkpoint_id)
    except NotFoundError as exc:
        secho_error_and_exit(str(exc))

    typer.secho(f"Checkpoint({checkpoint_id}) is successfully restored.")


# @app.command("import")
@check_api
def import_from_catalog(
    catalog_name: str = typer.Argument(
        ...,
        help="The name of public checkpoint to try out.",
    ),
    name: str = typer.Option(
        None,
        "--name",
        "-n",
        help="The name of the checkpoint that will be created in the project.",
    ),
    method: CatalogImportMethod = typer.Option(
        CatalogImportMethod.COPY.value,
        "--method",
        "-m",
        help=(
            "The method to import the public checkpoint. When the method is 'copy', "
            "the file objects will be copied to a separate storage so that the "
            "checkpoint remains available even if the source public checkpoint is "
            "removed from the catalog. Conversely, when the 'ref' method is selected, "
            "the file objects are referenced from the storage of the source public "
            "checkpoint. In this case, the imported checkpoint becomes unavailable if "
            "the source in the catalog is deleted."
        ),
    ),
):
    """Create a checkpoint by importing a public checkpoint in the catalog."""
    catalog_client = Catalog()
    catalogs = catalog_client.list(name=catalog_name)
    catalog = None
    for cat in catalogs:
        if cat.name == catalog_name:
            catalog = cat
            break

    if catalog is None:
        msg = (
            f"Public checkpoint with name '{catalog_name}' is not found in the catalog."
        )
        if len(catalogs) > 0:
            msg += f" Did you mean '{catalogs[0].name}'?"
        secho_error_and_exit(msg)

    if name is None:
        hash = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))
        name = f"{catalog.name}-{hash}"

    ckpt_client = Checkpoint()
    ckpt = ckpt_client.import_from_catalog(id=catalog.id, name=name, method=method)
    ckpt_dict = ckpt.model_dump()

    ckpt_dict["created_at"] = datetime_to_pretty_str(ckpt.created_at)
    status = get_translated_checkpoint_status(ckpt)
    ckpt_dict["status"] = status

    panel_formatter.render([ckpt_dict])
    json_formatter.render(ckpt_dict["attributes"])
    tree_formatter.render(ckpt_dict["forms"][0]["files"])


@app.command()
def convert(
    model_name_or_path: str = typer.Option(
        ...,
        "--model-name-or-path",
        "-m",
        help="Hugging Face pretrained model name or path to the saved model checkpoint.",
    ),
    output_dir: str = typer.Option(
        ...,
        "--output-dir",
        "-o",
        help=(
            "Directory path to save the converted checkpoint and related configuration "
            "files. Three files will be created in the directory: `model.h5`, "
            "`tokenizer.json`, and `attr.yaml`. "
            "The `model.h5` or `model.safetensors` is the converted checkpoint and can be renamed using "
            "the `--output-model-filename` option. "
            "The `tokenizer.json` is the Friendli-compatible tokenizer file, which should "
            "be uploaded along with the checkpoint file to tokenize the model input "
            "and output. "
            "The `attr.yaml` is the checkpoint attribute file, to be used when uploading "
            "the converted model to Friendli. You can designate the file name using "
            "the `--output-attr-filename` option."
        ),
    ),
    data_type: CheckpointDataType = typer.Option(
        ..., "--data-type", "-dt", help="The data type of converted checkpoint."
    ),
    cache_dir: Optional[str] = typer.Option(
        None, "--cache-dir", help="Directory for downloading checkpoint."
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Only check conversion avaliability."
    ),
    output_model_file_name: str = typer.Option(
        None,
        "--output-model-filename",
        help="Name of the converted checkpoint file."
        "The default file name is `model.h5` when `--output-ckpt-file-type` is `hdf5` or `model.safetensors` when `--output-ckpt-file-type` is `safetensors`.",
    ),
    output_ckpt_file_type: CheckpointFileType = typer.Option(
        CheckpointFileType.HDF5,
        "--output-ckpt-file-type",
        help="File format of the converted checkpoint file.",
    ),
    output_attr_file_name: str = typer.Option(
        "attr.yaml",
        "--output-attr-filename",
        help="Name of the checkpoint attribute file.",
    ),
    quantize: bool = typer.Option(
        False,
        "--quantize",
        help="Quantize the model before conversion",
    ),
    quant_config_file: Optional[typer.FileText] = typer.Option(
        None,
        "--quant-config-file",
        help="Path to the quantization configuration file.",
    ),
):
    """Convert huggingface's model checkpoint to Friendli format.

    When a checkpoint is in the Hugging Face format, it cannot be directly served.
    It requires conversion to the Friendli format for serving. The conversion
    process involves copying the original checkpoint and transforming it into a
    checkpoint in the Friendli format (*.h5).

    :::caution
    The `friendli checkpoint convert` is available only when the package is installed with
    `pip install "friendli-client[mllib]"`.
    :::

    ### Apply quantization

    If you want to quantize the model along with the conversion, `--quantize` option
    should be provided. You can customize the quantization configuration by describing
    it in a YAML file and providing the path to the file to `--quant-config-file`
    option. When `--quantize` option is used without providing `--quant-config-file`,
    the following configuration is used by default.

    ```yaml
    # Default quantization configuration
    mode: awq
    device: cuda:0
    seed: 42
    offload: true
    calibration_dataset:
        path_or_name: lambada
        format: json
        split: validation
        lookup_column_name: text
        num_samples: 128
        max_length: 512
    awq_args:
        quant_bit: 4
        quant_group_size: 64
    ```

    - **`mode`**: Quantization scheme to apply. Defaults to "awq".
    - **`device`**: Device to run the quantization process. Defaults to "cuda:0".
    - **`seed`**: Random seed. Defaults to 42.
    - **`offload`**: When enabled, this option significantly reduces GPU memory usage by offloading model layers onto CPU RAM. Defaults to true.
    - **`calibration_dataset`**
        - **`path_or_name`**: Path or name of the dataset. Datasets from either the Hugging Face Datasets Hub or local file system can be used. Defaults to "lambada".
        - **`format`**: Format of datasets. Defaults to "json".
        - **`split`**: Which split of the data to load. Defaults to "validation".
        - **`lookup_column_name`**: The name of a column in the dataset to be used as calibration inputs. Defaults to "text".
        - **`num_samples`**: The number of dataset samples to use for calibration. Note that the dataset will be shuffled before sampling. Defaults to 512.
        - **`max_length`**: The maximum length of a calibration input sequence. Defauts to 512.
    - **`awq_args`** (Fill in this field only for "awq" mode)
        - **`quant_bit`** : Bit width of integers to represent weights. Possible values are `4` or `8`. Defaults to 4.
        - **`quant_group_size`**: Group size of quantized matrices. 64 is the only supported value at this time. Defaults to 64.

    :::tip
    If you encounter OOM issues when running with AWQ, try enabling the `offload` option.
    :::

    :::tip
    If you set `percentile` in quant-config-file into 100,
    the quantization range will be determined by the maximum absolute values of the activation tensors.
    :::

    :::info
    Currently, [AWQ](https://arxiv.org/abs/2306.00978) is the only supported quantization scheme.
    :::

    :::info
    AWQ is supported only for models with architecture listed as follows:

    - `GPTNeoXForCausalLM`
    - `GPTJForCausalLM`
    - `LlamaForCausalLM`
    - `MPTForCausalLM`
    :::

    """
    try:
        from friendli.modules.converter.convert import (  # pylint: disable=import-outside-toplevel
            convert_checkpoint,
        )
        from friendli.modules.quantizer.schema.config import (  # pylint: disable=import-outside-toplevel
            AWQConfig,
            OneOfQuantConfig,
            QuantConfig,
        )
    except ModuleNotFoundError as exc:
        secho_error_and_exit(str(exc))

    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            secho_error_and_exit(f"'{output_dir}' exists, but it is not a directory.")
        os.mkdir(output_dir)

    quant_config: Optional[OneOfQuantConfig] = None
    if quantize:
        if quant_config_file:
            try:
                quant_config_dict = cast(dict, yaml.safe_load(quant_config_file.read()))
            except yaml.YAMLError as err:
                secho_error_and_exit(f"Failed to load the quant config file: {err}")
            quant_config = QuantConfig.model_validate(
                {"config": quant_config_dict}
            ).config
        else:
            quant_config = AWQConfig()

    default_names = {
        CheckpointFileType.HDF5: "model.h5",
        CheckpointFileType.SAFETENSORS: "model.safetensors",
    }
    output_model_file_name = (
        output_model_file_name or default_names[output_ckpt_file_type]
    )

    model_output_path = os.path.join(output_dir, output_model_file_name)
    tokenizer_output_dir = output_dir
    attr_output_path = os.path.join(output_dir, output_attr_file_name)

    try:
        convert_checkpoint(
            model_name_or_path=model_name_or_path,
            model_output_path=model_output_path,
            data_type=data_type,
            output_ckpt_file_type=output_ckpt_file_type,
            tokenizer_output_dir=tokenizer_output_dir,
            attr_output_path=attr_output_path,
            cache_dir=cache_dir,
            dry_run=dry_run,
            quantize=quantize,
            quant_config=quant_config,
        )
    except (NotFoundError, CheckpointConversionError, InvalidConfigError) as exc:
        secho_error_and_exit(str(exc))

    msg = (
        f"Checkpoint({model_name_or_path}) can be converted."
        if dry_run
        else f"Checkpoint({model_name_or_path}) has been converted successfully."
    )
    typer.secho(msg)


@app.command()
def convert_adapter(
    adapter_name_or_path: str = typer.Option(
        ...,
        "--adapter-name-or-path",
        "-a",
        help="Hugging Face pretrained adapter name or path to the saved adapter checkpoint.",
    ),
    output_dir: str = typer.Option(
        ...,
        "--output-dir",
        "-o",
        help=(
            "Directory path to save the converted adapter checkpoint and related configuration "
            "files. Two files will be created in the directory: `adapter.h5`, "
            "and `attr.yaml`. "
            "The `adapter.h5` is the converted checkpoint and can be renamed using "
            "the `--output-adapter-filename` option. "
            "The `attr.yaml` is the adapter checkpoint attribute file, to be used when uploading "
            "the converted model to Friendli. You can designate the file name using "
            "the `--output-attr-filename` option."
        ),
    ),
    data_type: CheckpointDataType = typer.Option(
        ..., "--data-type", "-dt", help="The data type of converted checkpoint."
    ),
    cache_dir: Optional[str] = typer.Option(
        None, "--cache-dir", help="Directory for downloading checkpoint."
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Only check conversion avaliability."
    ),
    output_adapter_filename: str = typer.Option(
        "adapter.h5",
        "--output-adapter-filename",
        help="Name of the converted adapter checkpoint file.",
    ),
    output_attr_filename: str = typer.Option(
        "adapter_attr.yaml",
        "--output-attr-filename",
        help="Name of the adapter checkpoint attribute file.",
    ),
) -> None:
    """Convert huggingface's adapter checkpoint to Friendli format.

    When an adapter checkpoint is in the Hugging Face PEFT format, it cannot
    be directly served in Friendli. It requires conversion to the Friendli format.
    The conversion process involves copying the original adapter checkpoint and
    transforming it into a checkpoint in the Friendli format (*.h5).

    This function does not include the `friendli checkpoint convert` command. i.e.
    `friendli checkpoint convert-adapter` only converts adapter's parameters, not backbone's.

    :::caution
    The `friendli checkpoint convert-adapter` is available only when the package is installed with
    `pip install "friendli-client[mllib]"`.
    :::

    """
    try:
        from friendli.modules.converter.convert import (  # pylint: disable=import-outside-toplevel
            convert_adapter_checkpoint,
        )
    except ModuleNotFoundError as exc:
        secho_error_and_exit(str(exc))

    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            secho_error_and_exit(f"'{output_dir}' exists, but it is not a directory.")
        os.mkdir(output_dir)

    # Engine cannot load a Safetensors Lora ckpt yet.
    output_adapter_file_type = CheckpointFileType.HDF5
    default_names = {
        CheckpointFileType.HDF5: "adapter.h5",
        CheckpointFileType.SAFETENSORS: "adapter.safetensors",
    }
    output_adapter_filename = (
        output_adapter_filename or default_names[output_adapter_file_type]
    )

    adapter_output_path = os.path.join(output_dir, output_adapter_filename)
    attr_output_path = os.path.join(output_dir, output_attr_filename)

    try:
        convert_adapter_checkpoint(
            adapter_name_or_path=adapter_name_or_path,
            adapter_output_path=adapter_output_path,
            adapter_attr_output_path=attr_output_path,
            data_type=data_type,
            output_adapter_file_type=output_adapter_file_type,
            cache_dir=cache_dir,
            dry_run=dry_run,
        )
    except (NotFoundError, CheckpointConversionError, InvalidConfigError) as exc:
        secho_error_and_exit(str(exc))



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli Credential SDK."""

# pylint: disable=line-too-long, arguments-differ, too-many-arguments, redefined-builtin

from __future__ import annotations

from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import ValidationError

from friendli.client.credential import CredentialClient
from friendli.client.project import ProjectCredentialClient
from friendli.enums import CredType
from friendli.errors import InvalidConfigError, NotSupportedError
from friendli.schema.resource.v1.credential import V1Credential
from friendli.sdk.resource.base import ResourceAPI
from friendli.utils.maps import cred_schema_map, cred_type_map_inv


class Credential(ResourceAPI[V1Credential, UUID]):
    """Friendli credential resource API."""

    def create(
        self, *, cred_type: CredType, name: str, value: Dict[str, Any]
    ) -> V1Credential:
        """Creates a new credential.

        Args:
            cred_type (CredType): The type of credential.
            name (str): The name of credential to create.
            value (Dict[str, Any]): Values of credential in plaintext.

        Raises:
            InvalidConfigError: Raised if `value` does not match the credential schema.

        Returns:
            V1Credential: The created credential object.

        """
        cred_schema = cred_schema_map[cred_type]
        try:
            cred_schema.model_validate(obj=value)
        except ValidationError as exc:
            raise InvalidConfigError(
                detail=f"Wrong credential format: {str(exc)}"
            ) from exc

        client = ProjectCredentialClient()
        raw_cred = client.create_credential(cred_type, name, 1, value)
        raw_cred["type"] = cred_type_map_inv[raw_cred["type"]].value
        cred = V1Credential.model_validate(raw_cred)
        return cred

    def get(self, id: UUID, *args, **kwargs) -> V1Credential:
        """[skip-doc] Gets a credential info.

        Args:
            id (UUID): ID of credential to retrieve.

        Returns:
            Dict[str, Any]: Retrieved credential info.

        """
        msg = "Getting a specific credential info is not supported."
        raise NotSupportedError(msg)

    def list(self, *, cred_type: Optional[CredType] = None) -> List[V1Credential]:
        """Lists credentials.

        Args:
            cred_type (Optional[CredType], optional): Filters by credential types. Defaults to None.

        Returns:
            List[V1Credential]: A list of retrieved credentials.

        """
        client = ProjectCredentialClient()
        raw_creds = client.list_credentials(cred_type)
        creds = []
        for raw_cred in raw_creds:
            cred = V1Credential.model_validate(raw_cred)
            cred_type = cred_type_map_inv.get(cred.type)
            if cred_type is None:
                continue
            cred.type = cred_type.value
            creds.append(cred)
        return creds

    def edit(
        self,
        id: UUID,
        *,
        name: Optional[str] = None,
        value: Optional[Dict[str, Any]] = None,
    ) -> V1Credential:
        """Edits credential info.

        Args:
            id (UUID): ID of credential to edit.
            name (Optional[str], optional): Name to update. No update if it is `None`. Defaults to None.
            value (Optional[Dict[str, Any]], optional): The new value of credential in plaintext. No update if it is `None`. Defaults to None.

        Returns:
            V1Credential: Updated credential object.

        """
        client = CredentialClient()
        raw_cred = client.update_credential(
            id, name=name, type_version="1", value=value
        )
        cred = V1Credential.model_validate(raw_cred)
        cred.type = cred_type_map_inv[cred.type].value
        return cred

    def delete(self, id: UUID) -> None:
        """Deletes a credential.

        Args:
            id (UUID): ID of credential to delete.

        """
        client = CredentialClient()
        client.delete_credential(id)



# Copyright (c) 2023-present, FriendliAI Inc. All rights reserved.

"""Friendli Client."""

from __future__ import annotations

from typing import Optional

import friendli
from friendli.sdk.api.chat.chat import AsyncChat, Chat
from friendli.sdk.api.completions import AsyncCompletions, Completions

SERVERLESS_ENDPOINT_URL = "https://inference.friendli.ai"


class FriendliClientBase:
    """Base class of Friendli client."""

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
    ):
        """Initializes FriendliClientBase."""
        if token is not None:
            friendli.token = token
        if team_id is not None:
            friendli.team_id = team_id


class Friendli(FriendliClientBase):
    """Friendli API client."""

    completions: Completions
    chat: Chat

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
    ):
        """Initializes Friendli."""
        super().__init__(
            token=token,
            team_id=team_id,
        )

        endpoint = SERVERLESS_ENDPOINT_URL
        self.completions = Completions(endpoint=endpoint)
        self.chat = Chat(endpoint=endpoint)


class AsyncFriendli(FriendliClientBase):
    """Async Friendli API client."""

    completions: AsyncCompletions
    chat: AsyncChat

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        team_id: Optional[str] = None,
    ):
        """Initializes AsyncFriendli."""
        super().__init__(token=token, team_id=team_id)

        endpoint = SERVERLESS_ENDPOINT_URL
        self.completions = AsyncCompletions(endpoint=endpoint)
        self.chat = AsyncChat(endpoint=endpoint)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Initialize Friendli."""

# pylint: disable=line-too-long

from __future__ import annotations

import friendli
from friendli.client.project import ProjectClient, find_project_id
from friendli.client.user import UserGroupClient, UserGroupProjectClient
from friendli.errors import AuthorizationError


def init(token: str, project: str) -> None:
    """Setup Friendli authorization info.

    All Friendli APIs followed by this function will be executed with the API key,
    provided in the `token` argument, and project context provided in the `project`
    argument. If you call the Friendli APIs without calling this function, the context
    set by `friendli login` CLI command will be used.

    Args:
        token (str): Friendli API key.
        project (str): Project name.

    Raises:
        AuthorizationError: Raised if the API key, provided in the `token` argument, does not have permission to any organization or to the project provided in the `project` argument.

    """
    friendli.token = token

    user_group_client = UserGroupClient()

    try:
        org = user_group_client.get_group_info()
    except IndexError as exc:
        raise AuthorizationError(
            "Does have not permission to any organization."
        ) from exc
    friendli.team_id = org["id"]

    user_group_project_client = UserGroupProjectClient()
    project_client = ProjectClient()

    project_id = find_project_id(
        projects=user_group_project_client.list_projects(),
        project_name=project,
    )
    if project_client.check_project_membership(pf_project_id=project_id):
        friendli.project_id = str(project_id)
    else:
        raise AuthorizationError(
            f"Does not have permission to the project '{project}'."
        )



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Friendli deployment SDK."""

# pylint: disable=line-too-long, arguments-differ, too-many-arguments, too-many-locals, redefined-builtin

from __future__ import annotations

import json
import os
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional
from uuid import UUID

from dateutil.tz import tzlocal
from pydantic import AnyHttpUrl

from friendli.client.deployment import (
    DeploymentClient,
    DeploymentEventClient,
    DeploymentLogClient,
    DeploymentMetricsClient,
    DeploymentReqRespClient,
    PFSProjectUsageClient,
)
from friendli.client.file import FileClient, GroupProjectFileClient
from friendli.client.group import GroupClient
from friendli.configurator.deployment import DRCConfigurator, OrcaDeploymentConfigurator
from friendli.context import get_current_group_id, get_current_project_id
from friendli.enums import (
    CloudType,
    DeploymentSecurityLevel,
    DeploymentType,
    GpuType,
    ServiceTier,
)
from friendli.errors import (
    AuthenticationError,
    EntityTooLargeError,
    InvalidConfigError,
    InvalidPathError,
    LowServicePlanError,
    NotFoundError,
)
from friendli.logging import logger
from friendli.schema.resource.v1.deployment import V1Deployment
from friendli.schema.resource.v1.transfer import UploadTask
from friendli.sdk.resource.base import ResourceAPI
from friendli.utils.format import extract_datetime_part, extract_deployment_id_part
from friendli.utils.maps import cloud_gpu_map, gpu_num_map
from friendli.utils.transfer import (
    MAX_DEFAULT_REQUEST_CONFIG_SIZE,
    ChunksizeAdjuster,
    DeferQueue,
    DownloadManager,
    UploadManager,
)
from friendli.utils.validate import validate_enums


class Deployment(ResourceAPI[V1Deployment, str]):
    """Deployment resource API."""

    def create(
        self,
        *,
        checkpoint_id: UUID,
        name: str,
        cloud: CloudType,
        region: str,
        gpu_type: GpuType,
        num_gpus: int,
        config: Dict[str, Any],
        deployment_type: DeploymentType = DeploymentType.PROD,
        description: Optional[str] = None,
        default_request_config: Optional[Dict[str, Any]] = None,
        security_level: DeploymentSecurityLevel = DeploymentSecurityLevel.PUBLIC,
        logging: bool = False,
        min_replicas: int = 1,
        max_replicas: int = 1,
    ) -> V1Deployment:
        """Creates a new deployment.

        Args:
            checkpoint_id (UUID): ID of checkpoint to deploy.
            name (str): The name of deployment.
            cloud (CloudType): Type of cloud provider.
            region (str): Cloud region to create a deployment.
            gpu_type (GpuType): Type of GPU.
            num_gpus (int): The number of GPUs.
            vm_type (VMType): Type of VM.
            config (Dict[str, Any]): Deployment configuration.
            deployment_type (DeploymentType, optional): Type of deployment. Defaults to DeploymentType.PROD.
            description (Optional[str], optional): Optional long description for the deployment. Defaults to None.
            default_request_config (Optional[Dict[str, Any]], optional): Default request configuration (e.g., stop words, bad words). Defaults to None.
            security_level (DeploymentSecurityLevel, optional): Security level of deployment endpoint. Defaults to DeploymentSecurityLevel.PUBLIC.
            logging (bool, optional): When True, enables request-response logging for the deployment if it is set. Defaults to False.
            min_replicas (int, optional): The number of minimum replicas. Defaults to 1.
            max_replicas (int, optional): The number of maximum replicas. Defaults to 1.

        Raises:
            AuthenticationError: Raised when project context is not configured.
            InvalidConfigError: Raised when any of the configurations (i.e., `config`, `default_request_config`, `min_replicas`, `max_replicas`) is invalid.
            EntityTooLargeError: Raised when the contents of `default_request_config` exceeds the 10GiB size limit.
            LowServicePlanError: Raised when the `deployment_type` is set to `Deployment.DEV` and service plan of the user's organization is not 'enterprise'.

        Examples:
            Basic usage:

            ```python
            from friendli import FriendliResource

            # Set up Friendli context.
            client = FriendliResource(
                token="YOUR_FRIENDLI_TOKEN",
                project_name="my-project",
            )

            # Create a deployment at GCP asia-northest3 region wtih one A100 GPU.
            config = {
                "max_batch_size": 256,
                "max_token_count": 8146,
                "max_num_tokens_to_replace": 0,
            }
            deployment = friendli.deployment.create(
                checkpoint_id="YOUR_CHECKPOINT_ID",
                name="my-deployment",
                cloud="gcp",
                region="asia-northeast3",
                gpu_type="a100",
                num_gpus=1,
                config=config,
            )
            ```

            The format of `config` should be:

            ```python
            {
                "max_batch_size": Optioanl[int],
                "max_token_count": Optioanl[int],
                "max_num_tokens_to_replace": Optional[int],
            }
            ```

            The format of `default_request_config` should be:

            ```python
            {
                "stop": Optional[List[str]],
                "stop_tokens": Optional[List[int]],
                "bad_words": Optional[List[str]],
                "bad_word_tokens": Optional[List[int]]
            }
            ```

            :::caution
            Note that `bad_words` and `bad_word_tokens` cannot be set at the same time. Similarly, `stop` and `stop_tokens` cannot be set at the same time.
            :::

            :::note
            When `min_replicas` and `max_replicas` are the same, deployment auto-scaling turns off.
            :::

        Returns:
            V1Deployment: The created deployment object.

        """
        # pylint: disable=too-many-statements
        cloud = validate_enums(cloud, CloudType)
        gpu_type = validate_enums(gpu_type, GpuType)
        deployment_type = validate_enums(deployment_type, DeploymentType)
        security_level = validate_enums(security_level, DeploymentSecurityLevel)

        org_id = get_current_group_id()
        if org_id is None:
            raise AuthenticationError(
                "Not authenticated. Valid access token is not provided."
            )
        project_id = get_current_project_id()
        if project_id is None:
            raise AuthenticationError("Project context is not configured.")

        group_client = GroupClient()
        if (
            group_client.get_group(pf_group_id=org_id)["plan"] == ServiceTier.BASIC
            and deployment_type == DeploymentType.DEV
        ):
            raise LowServicePlanError(
                "Deployment with the development type is only supported for the 'enterprise' plan."
            )

        if min_replicas > max_replicas:
            raise InvalidConfigError(
                f"Should be min_replicas('{min_replicas}') <= max_replicas('{max_replicas}')."
            )

        if gpu_type not in cloud_gpu_map[cloud]:
            raise InvalidConfigError(
                f"GPU type {gpu_type.value} is not supported in cloud {cloud.value}."
            )

        if num_gpus not in gpu_num_map[gpu_type]:
            raise InvalidConfigError(
                f"Num gpus {num_gpus} is not supported for GPU {gpu_type.value}."
            )

        deploy_configurator = OrcaDeploymentConfigurator(config=config)
        deploy_configurator.validate()
        config = {"orca_config": config}

        if default_request_config is not None:
            drc_configurator = DRCConfigurator(config=default_request_config)
            drc_configurator.validate()

            file_client = FileClient()
            group_file_client = GroupProjectFileClient()

            with TemporaryDirectory() as dir:
                drc_file_name = "drc.json"
                drc_file_path = os.path.join(dir, drc_file_name)
                with open(drc_file_path, "w", encoding="utf-8") as file:
                    json.dump(default_request_config, file)

                file_size = os.stat(drc_file_path).st_size
                if file_size > MAX_DEFAULT_REQUEST_CONFIG_SIZE:
                    raise EntityTooLargeError(
                        "The default request config size should be smaller than 1GiB."
                    )

                file_info = {
                    "name": drc_file_name,
                    "path": drc_file_name,
                    "mtime": datetime.fromtimestamp(
                        os.stat(drc_file_path).st_mtime, tz=tzlocal()
                    ).isoformat(),
                    "size": file_size,
                }
                file_id = group_file_client.create_misc_file(file_info=file_info)["id"]

                upload_url = file_client.get_misc_file_upload_url(misc_file_id=file_id)
                upload_task = UploadTask(
                    path=drc_file_path, upload_url=AnyHttpUrl(upload_url)
                )

                executor = ThreadPoolExecutor()
                adjuster = ChunksizeAdjuster()
                upload_manager = UploadManager(
                    executor=executor, chunk_adjuster=adjuster
                )
                try:
                    upload_manager.upload_file(upload_task=upload_task)
                    file_client.make_misc_file_uploaded(misc_file_id=file_id)
                    config["orca_config"]["default_request_config_id"] = file_id
                finally:
                    executor.shutdown(wait=True)

        config["orca_config"]["num_devices"] = num_gpus

        config["scaler_config"] = {}
        config["scaler_config"]["min_replica_count"] = min_replicas
        config["scaler_config"]["max_replica_count"] = max_replicas

        request_data = {
            "project_id": str(project_id),
            "model_id": str(checkpoint_id),
            "deployment_type": deployment_type.value,
            "name": name,
            "vm": {"gpu_type": gpu_type.value},
            "cloud": cloud.value,
            "region": region,
            "total_gpus": num_gpus,
            "infrequest_perm_check": security_level
            == DeploymentSecurityLevel.PROTECTED,
            "infrequest_log": logging,
            **config,
        }
        if description is not None:
            request_data["description"] = description

        client = DeploymentClient()
        deployment_raw = client.create_deployment(request_data)
        deployment = V1Deployment.model_validate(deployment_raw)
        return deployment

    def list(
        self,
        *,
        limit: int = 20,
        include_terminated: bool = False,
        from_oldest: bool = False,
        all_in_org: bool = False,
    ) -> List[V1Deployment]:
        """Lists deployments.

        Args:
            limit (int, optional): The maximum number of retrieved results. Defaults to 20.
            include_terminated (bool, optional): When True, includes the terminated deployments in the results. Defaults to False.
            from_oldest (bool, optional): List from the oldest deployment. Defaults to False.

        Raises:
            AuthenticationError: Raised when project context is not configured.

        Returns:
            List[V1Deployment]: Retrieved deployments.

        """
        if all_in_org:
            prj_id = None
        else:
            project_id = get_current_project_id()
            if project_id is None:
                raise AuthenticationError("Project context is not configured.")
            prj_id = str(project_id)

        client = DeploymentClient()
        deployments = [
            V1Deployment.model_validate(deployment_raw)
            for deployment_raw in client.list_deployments(
                project_id=prj_id,
                archived=False,
                limit=limit,
                from_oldest=from_oldest,
            )
        ]
        num_active_deployments = len(deployments)
        if include_terminated and limit > num_active_deployments:
            deployments += [
                V1Deployment.model_validate(deployment_raw)
                for deployment_raw in client.list_deployments(
                    project_id=prj_id,
                    archived=True,
                    limit=limit - num_active_deployments,
                    from_oldest=from_oldest,
                )
            ]

        return deployments

    def get(self, id: str, *args, **kwargs) -> V1Deployment:
        """Gets deployment info.

        Args:
            id (str): ID of deployment to retrieve.

        Returns:
            V1Deployment: Retrieved deployment object.

        """
        client = DeploymentClient()
        deployment_raw = client.get_deployment(id)
        deployment = V1Deployment.model_validate(deployment_raw)

        return deployment

    def stop(self, id: str) -> None:
        """Stops a running deployment.

        Args:
            id (str): ID of deployment to stop.

        """
        client = DeploymentClient()
        client.stop_deployment(id)

    def get_metrics(
        self,
        id: str,
        *,
        since: datetime,
        until: datetime,
        time_window: int = 60,
    ) -> List[Dict[str, Any]]:
        """Gets metrics of a deployment.

        Args:
            id (str): ID of deployment to get metrics.
            since (datetime): Start time of metrics to fetch.
            until (datetime): End time of metrics to fetch.
            time_window (int, optional): Time window of results in seconds. Defaults to 60.

        Returns:
            List[Dict[str, Any]]: Retrieved metrics data.

        """
        metrics_client = DeploymentMetricsClient(deployment_id=id)
        return metrics_client.get_metrics(
            start=since, end=until, time_window=time_window
        )

    def get_project_deployment_durations(
        self, *, since: datetime, until: datetime
    ) -> Dict[str, Any]:
        """Gets usage info of a deployment.

        Args:
            since (datetime): Start datetime of the deployment usages to fetch.
            until (datetime): End datetime of the deployment usages to fetch.

        Returns:
            Dict[str, Any]: Retrieved deployment usage info.

        """
        client = PFSProjectUsageClient()
        return client.get_project_deployment_durations(since, until)

    def get_logs(self, id: str, *, replica_index: int = 0) -> List[Dict[str, Any]]:
        """Gets logs from a deployment.

        Args:
            id (str): ID of deployment to get logs.
            replica_index (int, optional): Index of deployment replica to retrieve logs. Defaults to 0.

        Returns:
            List[Dict[str, Any]]: Retrieved log data.

        """
        client = DeploymentLogClient(deployment_id=id)
        return client.get_deployment_logs(replica_index=replica_index)

    def adjust_replica_config(
        self, id: str, *, min_replicas: int, max_replicas: int
    ) -> None:
        """Adjusts replica configuration of a running deployment.

        Args:
            id (str): ID of deployment to adjust the replica configuration.
            min_replicas (int): Minimum replica count.
            max_replicas (int): Maximum replica count.

        Raises:
            InvalidConfigError: Raised when any of the configurations (i.e., `min_replicas`, `max_replicas`) is invalid.

        """
        if min_replicas > max_replicas:
            raise InvalidConfigError(
                f"Should be min_replicas('{min_replicas}') <= max_replicas('{max_replicas}')."
            )
        client = DeploymentClient()
        client.update_deployment_scaler(
            deployment_id=id,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
        )

    def get_events(self, id: str) -> List[Dict[str, Any]]:
        """Gets events from a deployment.

        Args:
            id (str): ID of deployment to get events.

        Returns:
            List[Dict[str, Any]]: Deployment event info.

        """
        client = DeploymentEventClient(deployment_id=id)
        return client.get_events()

    def download_req_resp_logs(
        self,
        id: str,
        *,
        since: datetime,
        until: datetime,
        save_dir: Optional[str] = None,
    ) -> None:
        """Downloads a file that has request-response logs.

        Args:
            id (str): ID of deployment.
            since (datetime): Start time of request-response logs.
            until (datetime): End time of request-response logs.
            save_dir (Optional[str], optional): Path to save logs. Set to the current directory if it is `None`. Defaults to None.

        Raises:
            InvalidPathError: Raised when `save_dir` does not exist or is read-only.
            InvalidConfigError: Raised when `since` <= `until` is not satisfied.
            NotFoundError: Raised when the deployment request-response logs are not found for the given time range.

        """
        if save_dir is not None and not os.path.isdir(save_dir):
            raise InvalidPathError(f"Directory '{save_dir}' is not found.")
        save_dir = save_dir or os.getcwd()

        if not os.access(save_dir, os.W_OK):
            raise InvalidPathError(f"Cannot save logs to {save_dir} which is readonly.")

        if since > until:
            raise InvalidConfigError(
                "Time value of `since` should be earlier than the value of `until`."
            )

        client = DeploymentReqRespClient(deployment_id=id)
        download_infos = client.get_download_urls(start=since, end=until)
        if len(download_infos) == 0:
            raise NotFoundError(f"No log exists for the deployment '{id}'.")

        write_queue = DeferQueue()
        download_manager = DownloadManager(write_queue=write_queue)
        for i, download_info in enumerate(download_infos):
            full_storage_path = download_info["path"]
            deployment_id_part = extract_deployment_id_part(full_storage_path)
            timestamp_part = extract_datetime_part(full_storage_path)
            filename = f"{deployment_id_part}_{timestamp_part}.log"
            logger.info("Downloading files %d/%d...", i + 1, len(download_infos))
            download_manager.download_file(
                url=download_info["url"], out=Path(save_dir) / filename
            )
